# Copyright (c) 2018-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import copy
import os

import ephem
import h5py
import numpy as np
import traitlets
from astropy import units as u
from scipy.interpolate import RectBivariateSpline
from scipy.signal import square
from toast import qarray as qa
from toast.data import Data
from toast.instrument import Focalplane
from toast.observation import default_values as defaults
from toast.ops.operator import Operator
from toast.timing import Timer, function_timer
from toast.traits import Float, Instance, Int, List, Quantity, Unicode, trait_docs
from toast.utils import Logger, unit_conversion

from sotodlib.coords import local

from . import utils


def source_spectrum(fc, width, power, gain, amplitude, diameter, noise_bw, noise_out):
    """Generate a power spectrum for a source.

    The source has a delta-like emission.

    Args:
        fc (Quantity): Central frequency of the source signal
        width (Quantity): total width of the source signal
        power (Float): total power of the source signal in dBm
        gain (Float): Antenna gain of the source in dBi
        amplitude (Float): amplitude of the signal with respect to the the base in dB
        diameter (Quantity): source diameter
        noise_bw (Quantity): in-band noise
        noise_out (Quantity): out-of-band noise
    Returns:
        I_nu (Quantity): spectrum of the source

    """

    amplitude = 10 ** (-amplitude / 10)
    power_watts = 10 ** (power / 10) / 1000

    freqs = np.linspace(20.0, 350.0, 1001, endpoint=True) * 1e9 * u.Hz
    fc = fc.to(u.Hz)
    width = width.to(u.Hz)
    radius = diameter.to(u.meter) / 2

    if np.amin(np.diff(freqs)) >= 2 * width:
        idx = np.argmin(np.abs(freqs - fc))

        idx = np.array([idx - 1, idx, idx + 1])

    else:
        fmin = fc - width
        fmax = fc + width

        idx_min = np.argmin(np.abs(freqs - fmin))
        idx_max = np.argmin(np.abs(freqs - fmax))

        idx = np.arange(idx_min, idx_max + 1, dtype=int)

    bandwidth = np.ptp(freqs[idx])

    rho_bw = power_watts / bandwidth / (1 + amplitude) * u.W

    power_bw = rho_bw * bandwidth

    out1 = np.ptp(freqs[: idx[0]])
    out2 = np.ptp(freqs[idx[-1] + 1 :])

    rho_out = power_bw * amplitude / (out1 + out2)

    signal = np.zeros(len(freqs)) * u.W / u.Hz

    if noise_bw.value != 0:
        rho_bw += np.random.normal(0, np.abs(noise_bw), len(rho_bw))

    if noise_out.value != 0:
        rho_out += np.random.normal(0, np.abs(noise_out), len(rho_out))

    signal[: idx[0]] = rho_out
    signal[idx] = rho_bw
    signal[idx[-1] + 1 :] = rho_out

    g = 10 ** (gain / 10)

    I_nu = signal * g / np.pi**2 / radius**2

    return freqs, I_nu


class SimulateDroneMovement:

    def __init__(self, params):
        """Class containing the necessary functions to simulate a drone scanning above the telescope.

        Args:
            params (Dictionary): Dictionary with the parameters required to generate the movement of the drone.
                                 Keys available on the dictionary:
                                    azimuth_starting (Quantity): Starting value for the azimuth
                                    azimuth_range (Quantity): Range of the azimuth scan
                                    azimuth_velocity (Quantity): Maximum velocity of the drone along the azimuthal axis
                                    azimuth_acceleration (Quantity): Maximum velocity of the drone along the azimuthal axis
                                    azimuth_direction (String): Determine if the azimuth value is increasing or decreasing.
                                                                Accepted values: 'increasing' or 'decresing'
                                    elevation_starting (Quantity): Starting value for the elevation
                                    elevation_range (Quantity): Range of the elevation scan
                                    elevation_velocity (Quantity): Maximum velocity of the drone along the elevational axis
                                    elevation_acceleration (Quantity): Maximum velocity of the drone along the elevational axis
                                    elevation_direction (String): Determine if the elevation value is increasing or decreasing.
                                                                  Accepted values: 'increasing' or 'decresing'
                                    elevation_step (Quantity): Step along the elevation axis for a grid scan
                                    scan_time (Quantity): Time of the scan
                                    scan_type (String): Type of the scan
                                                        Accepted values:
                                                            elevation_only
                                                            azimuth_only
                                                            elevation_only_single
                                                            azimuth_only_single
                                                            grid
                                                            fixed

        """

        self.params = params

    def _check_maximum_velocity(self, delta_axis, vel, acc, direction):
        """Internal Function to control if the drone achieves the maximum velocity along a specific axis.
        Args:
            delta_axis (Quantity): span of the scan
            vel (Quantity): maximum velocity achievable
            acc (Quantity): value of the acceleration
            direction (Float): value that indicates the direction of movement

        Returns:
            t_acc (Quantity): total time of the acceleration (deceleration) phase
            t_vel (Quantity): total time at the maximum velocity achivable (0 if v_max < vel)
            v_max (Quantity): maximum velocity of the drone

        """

        if (
            self.params["scan_type"] == "elevation_only_single"
            or self.params["scan_type"] == "azimuth_only_single"
        ):
            v_max = (
                acc * self.params["scan_time"]
                - np.sqrt((acc * self.params["scan_time"]) ** 2 - 4 * acc * delta_axis)
            ) / 2
            t_acc = np.abs(v_max / acc)
            t_vel = self.params["scan_time"] - 2 * t_acc
        else:

            fact = delta_axis - np.abs(vel**2 / acc)

            if fact > 0:
                t_acc = np.abs(vel / acc)
                t_vel = (delta_axis - np.abs(vel**2 / acc)) / np.abs(vel)
                v_max = vel
            else:
                v_max = np.sqrt(np.abs(delta_axis * acc)) * direction
                t_acc = np.abs(v_max / acc)
                t_vel = 0

        return t_acc, t_vel, v_max

    def create_scan(self, time_interp=None):
        """Calculate the movement of the drone using the parameters dictionary.
        Args:
            times_interp (Quantity): times where to compute the position using a linear interpolation

        Return:
            time (Quantity): total duration of the scan
            azimuth (Quantity): positions during the scan along the azimuth axis
            elevation (Quantity): positions during the scan along the elevation axis
        """

        if self.params["scan_type"] == "elevation_only":
            time, azimuth, elevation = self.single_axis_scan(axis="elevation")
        elif self.params["scan_type"] == "azimuth_only":
            time, azimuth, elevation = self.single_axis_scan(axis="azimuth")
        elif self.params["scan_type"] == "elevation_only_single":

            if self.params["elevation_direction"].lower() == "decreasing":
                direction = -1
            else:
                direction = 1

            time, elevation = self.single_movement(
                axis="elevation",
                direction=direction,
                axis0=self.params["elevation_starting"],
            )

            azimuth = np.ones(len(elevation)) * self.params["azimuth_starting"]

        elif self.params["scan_type"] == "azimuth_only_single":

            if self.params["azimuth_direction"].lower() == "decreasing":
                direction = -1
            else:
                direction = 1

            time, azimuth = self.single_movement(
                axis="azimuth",
                direction=direction,
                axis0=self.params["azimuth_starting"],
            )

            elevation = np.ones(len(azimuth)) * self.params["elevation_starting"]

        elif self.params["scan_type"] == "grid_scan":

            time, azimuth, elevation = self.grid_scan()

        elif self.params["scan_type"] == "fixed":

            time = np.linspace(0, self.params["scan_time"], 1001, endpoint=True)
            azimuth = np.ones(len(time)) * self.params["azimuth_starting"]
            elevation = np.ones(len(time)) * self.params["elevation_starting"]

        if time_interp is not None:

            azimuth = np.interp(time_interp - time_interp[0], time, azimuth)
            elevation = np.interp(time_interp - time_interp[0], time, elevation)

            time = copy.copy(time_interp)

        return time, azimuth, elevation

    def single_movement(
        self, axis=None, direction=1, axis0=0, step=False, step_value=0
    ):
        """Calculate the movement along a specific axis for a single movement (i.e. up to down or down to up,
        left to right or right to left).
        Args:
            axis (String): axis along the movement needs to be generated
            direction (Float): value that indicates the direction of movement
            axis0 (Quantity): Starting point of the movement
            step (Bool): True is the movement is a stepping movement during a grid scan
            step_value (Quantity): range value of the step

        Return:
            time (Quantity): total duration of the single movement
            axis_val (Quantity): positions during the single movement
        """

        range_string = axis + "_range"
        vel_string = axis + "_velocity"
        acc_string = axis + "_acceleration"

        vel = self.params[vel_string] * direction
        acc = self.params[acc_string] * direction
        if step:
            delta_axis = step_value
        else:
            delta_axis = self.params[range_string]

        t_acc, t_vel, v_max = self._check_maximum_velocity(
            delta_axis, vel, acc, direction
        )

        axis_val = np.array([]) * u.rad
        time = np.array([]) * u.s

        time_acc = np.linspace(0, t_acc, 101, endpoint=True)
        time = np.append(time, time_acc)

        dx_acc = 0.5 * acc * time_acc**2
        axis_val = np.append(axis_val, axis0 + dx_acc)

        if t_vel > 0:
            time_vel = np.linspace(0, t_vel, 101, endpoint=True)
            time = np.append(time, time[-1] + time_vel[1:])

            dx_vel = v_max * time_vel[1:]
            axis_val = np.append(axis_val, axis_val[-1] + dx_vel)

        dx_dec = v_max * time_acc[1:] - 0.5 * acc * time_acc[1:] ** 2
        axis_val = np.append(axis_val, axis_val[-1] + dx_dec)

        time = np.append(time, time[-1] + time_acc[1:])

        return time, axis_val

    def single_axis_scan(self, axis="elevation"):
        """Calculate the movement along a specific axis for a scan (i.e. multiple up to down/down to up,
        multiple left to right/right to left).
        Args:
            axis (String): axis along the movement needs to be generated

        Return:
            time (Quantity): total duration of the scan
            azimuth (Quantity): positions during the scan along the azimuth axis
            elevation (Quantity): positions during the scan along the elevation axis
        """

        axis_value = np.array([]) * u.rad
        time = np.array([]) * u.s

        axis0_string = axis + "_starting"

        axis0 = copy.copy(self.params[axis0_string])
        t0 = 0

        if self.params[axis + "_direction"].lower() == "decreasing":
            direction = -1
        else:
            direction = 1

        self.scan_cycles = 0
        self._scan_idxs = np.array([], dtype=int)

        start = axis + "_starting"
        rng = axis + "_range"

        while True:
            time_temp, axis_temp = self.single_movement(
                axis=axis, direction=direction, axis0=axis0
            )
            axis0 = axis_temp[-1]
            direction *= -1

            time = np.append(time, time_temp + t0)
            axis_value = np.append(axis_value, axis_temp)
            t0 = copy.copy(time[-1])

            self.scan_cycles += 1

            if time[-1] > self.params["scan_time"]:

                idx = np.argmin(np.abs(time - self.params["scan_time"]))

                time = time[: idx + 1]
                axis_value = axis_value[: idx + 1]
                self._scan_idxs = np.append(self._scan_idxs, len(time) - 1)

                if direction == 1:
                    self.scan_cycles += (
                        axis_value[-1] - self.params[start]
                    ) / self.params[rng]
                elif direction == -1:
                    self.scan_cycles += (
                        (self.params[start] + self.params[rng] - axis_value[-1])
                    ) / self.params[rng]

                break

            self._scan_idxs = np.append(self._scan_idxs, len(time) - 1)

        self.scan_cycles /= 2
        self.scan_time = time
        if axis == "elevation":
            elevation = axis_value
            azimuth = np.ones(len(axis_value)) * self.params["azimuth_starting"]
        if axis == "azimuth":
            azimuth = axis_value
            elevation = np.ones(len(axis_value)) * self.params["elevation_starting"]

        return time, azimuth, elevation

    def grid_scan(self):
        """Calculate the movement during a grid scan with an elevation step.
        Args:
            None
        Return:
            time (Quantity): total duration of the scan
            azimuth (Quantity): positions during the scan along the azimuth axis
            elevation (Quantity): positions during the scan along the elevation axis
        """

        el_step = self.params["el_step"]

        az0 = self.params["azimuth_starting"]
        el0 = self.params["elevation_starting"]

        if self.params["azimuth_direction"].lower() == "decreasing":
            azimuth_direction = -1
        else:
            azimuth_direction = 1

        if self.params["elevation_direction"].lower() == "decreasing":
            elevation_direction = -1
        else:
            elevation_direction = 1

        el_cumulative = 0 * u.degree
        t0 = 0 * u.s

        time = np.array([]) * u.s
        azimuth_value = np.array([]) * u.rad
        elevation_value = np.array([]) * u.rad

        self.scan_cycles = 0
        self._scan_idxs = np.array([], dtype=int)

        while True:

            time_temp, axis_temp = self.single_movement(
                axis="azimuth", direction=azimuth_direction, axis0=az0, points=101
            )

            azimuth_direction *= -1
            az0 = copy.copy(axis_temp[-1])

            time = np.append(time, time_temp[1:] + t0)
            azimuth_value = np.append(azimuth_value, axis_temp[1:])
            elevation_value = np.append(
                elevation_value, np.ones_like(axis_temp[1:]) * el0
            )
            t0 = copy.copy(time[-1])

            if time[-1] > self.params["scan_time"]:

                idx = np.argmin(np.abs(time - self.params["scan_time"]))

                time = time[: idx + 1]
                azimuth_value = azimuth_value[: idx + 1]
                elevation_value = elevation_value[: idx + 1]

                break

            el_cumulative += el_step

            if el_cumulative >= self.params["elevation_range"]:
                time_temp, axis_temp = self.single_movement(
                    axis="elevation",
                    direction=elevation_direction,
                    axis0=el0,
                    step=True,
                    step_value=np.abs(self.params["elevation_range"] - el_cumulative),
                    points=101,
                )
                el_cumulative = 0
                elevation_direction *= -1
                self.scan_cycles += 1
                self._scan_idxs = np.append(
                    self._scan_idxs, len(time) + len(time_temp) - 1
                )
            else:
                time_temp, axis_temp = self.single_movement(
                    axis="elevation",
                    direction=elevation_direction,
                    axis0=el0,
                    step=True,
                    step_value=el_step,
                    points=101,
                )

            el0 = copy.copy(axis_temp[-1])
            time = np.append(time, time_temp[1:] + t0)
            elevation_value = np.append(elevation_value, axis_temp[1:])
            azimuth_value = np.append(azimuth_value, np.ones_like(axis_temp[1:]) * az0)

            t0 = copy.copy(time[-1])

            if time[-1] > self.params["scan_time"]:

                idx = np.argmin(np.abs(time - self.params["scan_time"]))

                time = time[: idx + 1]
                azimuth_value = azimuth_value[: idx + 1]
                elevation_value = elevation_value[: idx + 1]

                self._scan_idxs = np.append(self._scan_idxs, len(time) - 1)

                if elevation_direction == 1:
                    self.scan_cycles += (
                        elevation_value[-1] - self.params["elevation_starting"]
                    ) / self.params["elevation_range"]
                elif elevation_direction == -1:
                    self.scan_cycles += (
                        (
                            self.params["elevation_starting"]
                            + self.params["elevation_range"]
                            - elevation_value[-1]
                        )
                    ) / self.params["elevation_range"]

                break

        (idx,) = np.where(np.diff(time) != -1000)

        azimuth = azimuth_value
        elevation = elevation_value

        return time, azimuth, elevation


@trait_docs
class SimSource(Operator):
    """Operator that generates an Artificial Source timestreams."""

    # Class traits

    API = Int(0, help="Internal interface version for this operator")

    times = Unicode(
        defaults.times,
        help="Observation shared key for timestamps",
    )

    # Drone Position parameters

    source_distance = Quantity(
        500 * u.meter,
        help="Initial distance of the artificial source in meters",
    )

    source_azimuth_start = Quantity(
        u.Quantity(0, u.degree), help="Start of the scan along the azimuth axis"
    )

    source_azimuth_range = Quantity(
        u.Quantity(0, u.degree), help="Range of the scan along the azimuthal axis"
    )

    source_azimuth_velocity = Quantity(
        u.Quantity(2, u.Unit("deg / s")),
        help="Maximum velocity of the drone along the azimuthal axis",
    )

    source_azimuth_acceleration = Quantity(
        u.Quantity(2, u.Unit("deg / s2")),
        help="Maximum acceleration of the drone along the azimuthal axis",
    )

    source_azimuth_direction = Unicode(
        "increasing", help="Determine if the azimuth value is increasing or decreasing"
    )

    source_elevation_start = Quantity(
        u.Quantity(0, u.degree), help="Start of the scan along the elevation axis"
    )

    source_elevation_range = Quantity(
        u.Quantity(0, u.degree), help="Range of the scan along the elevation axis"
    )

    source_elevation_velocity = Quantity(
        u.Quantity(2, u.Unit("deg / s")),
        help="Maximum velocity of the drone along the elevation axis",
    )

    source_elevation_acceleration = Quantity(
        u.Quantity(2, u.Unit("deg / s2")),
        help="Maximum acceleration of the drone along the elevation axis",
    )

    source_elevation_direction = Unicode(
        "increasing",
        help="Determine if the elevation value is increasing or decreasing",
    )

    source_elevation_step = Quantity(
        u.Quantity(0, u.degree), help="Step along the elevation axis for a grid scan"
    )

    source_scan_type = Unicode("elevation_only", help="Type of the scan")

    source_err = List(
        [0, 0, 0],
        help="Source Position Error in ECEF Coordinates as [[X, Y, Z]] in meters",
    )

    source_size = Quantity(
        u.Quantity(0.1, u.meter),
        help="Source Size in meters",
    )

    source_power = Float(help="Max amplitude of the source in dBm")

    source_fc = Quantity(u.Quantity(90e9, u.Hz), help="Central frequency of the source")

    source_width = Quantity(u.Quantity(100e3, u.Hz), help="Width of the source signal")

    source_amplitude = Float(
        help="Amplitude of the source signal with the respect to the background in dB"
    )

    source_gain = Float(help="Gain of the source Antenna in dBi")

    source_noise_bw = Quantity(
        u.Quantity(0, u.W / u.Hz), help="White noise level in the emission band"
    )

    source_noise_out = Quantity(
        u.Quantity(0, u.W / u.Hz), help="White noise level outside the emission band"
    )

    source_freq_chopping = Quantity(
        u.Quantity(0, u.Hz), help="Frequency of the source chopping system"
    )

    source_pol_angle = Quantity(
        u.Quantity(90, u.degree),
        help="Angle of the polarization vector emitted by the source in degrees (0 means parallel to the gorund and 90 vertical)",
    )

    source_pol_angle_error = Quantity(
        u.Quantity(0, u.degree),
        help="Error in the angle of the polarization vector",
    )

    polarization_fraction = Float(
        1,
        help="Polarization fraction of the emitted signal",
    )

    beam_file = Unicode(
        None,
        allow_none=True,
        help="HDF5 file that stores the simulated beam",
    )

    # Drone parameters

    drone_temp = Quantity(u.Quantity(0, u.K), help="BlackBody temperature of the Drone")

    drone_emiss = Float(0, help="Emissivity of the drone")

    drone_size = Quantity(
        u.Quantity(0, u.meter),
        help="Drone size in meters",
    )

    # Wind Parameters

    wind_gusts_amp = Quantity(
        u.Quantity(0.0, u.Unit("m / s")), help="Amplitude of gusts of wind"
    )

    wind_gusts_duration = Quantity(
        u.Quantity(0.0, u.second), help="Duration of each gust of wind"
    )

    wind_gusts_number = Float(0, help="Number of wind gusts")

    wind_damp = Float(
        0, help="Dampening effect to reduce the movement of the drone due to gusts"
    )

    det_data = Unicode(
        defaults.det_data,
        help="Observation detdata key for simulated signal",
    )

    detector_pointing = Instance(
        klass=Operator,
        allow_none=True,
        help="Operator that translates boresight Az/El pointing into detector frame",
    )

    detector_weights = Instance(
        klass=Operator,
        allow_none=True,
        help="Operator that translates boresight Az/El pointing into detector weights",
    )

    elevation = Unicode(
        defaults.elevation,
        allow_none=True,
        help="Observation shared key for boresight elevation",
    )

    azimuth = Unicode(
        defaults.azimuth,
        allow_none=True,
        help="Observation shared key for azimuth",
    )

    @traitlets.validate("beam_file")
    def _check_beam_file(self, proposal):
        beam_file = proposal["value"]
        if beam_file is not None and not os.path.isfile(beam_file):
            raise traitlets.TraitError(f"{beam_file} is not a valid beam file")
        return beam_file

    @traitlets.validate("detector_pointing")
    def _check_detector_pointing(self, proposal):
        detpointing = proposal["value"]
        if detpointing is not None:
            if not isinstance(detpointing, Operator):
                raise traitlets.TraitError(
                    "detector_pointing should be an Operator instance"
                )
            # Check that this operator has the traits we expect
            for trt in [
                "view",
                "boresight",
                "shared_flags",
                "shared_flag_mask",
                "quats",
                "coord_in",
                "coord_out",
            ]:
                if not detpointing.has_trait(trt):
                    msg = f"detector_pointing operator should have a '{trt}' trait"
                    raise traitlets.TraitError(msg)
        return detpointing

    @traitlets.validate("detector_weights")
    def _check_detector_weights(self, proposal):
        detweights = proposal["value"]
        if detweights is not None:
            if not isinstance(detweights, Operator):
                raise traitlets.TraitError(
                    "detector_weights should be an Operator instance"
                )
            # Check that this operator has the traits we expect
            for trt in [
                "view",
                "weights",
                "mode",
            ]:
                if not detweights.has_trait(trt):
                    msg = f"detector_weights operator should have a '{trt}' trait"
                    raise traitlets.TraitError(msg)
        return detweights

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Store of per-detector beam properties.  Eventually we could modify the
        # operator traits to list files per detector, per wafer, per tube, etc.
        # For now, we use the same beam for all detectors, so this will have only
        # one entry.
        self.beam_props = dict()

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()
        comm = data.comm

        for trait in "beam_file", "detector_pointing":
            value = getattr(self, trait)
            if value is None:
                raise RuntimeError(f"You must set `{trait}` before running SimSource")

        timer = Timer()
        timer.start()

        if data.comm.group_rank == 0:
            log.debug(f"{data.comm.group} : Simulating Source")

        for obs in data.obs:
            observer = ephem.Observer()
            site = obs.telescope.site
            observer.lon = site.earthloc.lon.to_value(u.radian)
            observer.lat = site.earthloc.lat.to_value(u.radian)
            observer.elevation = site.earthloc.height.to_value(u.meter)

            prefix = f"{comm.group} : {obs.name}"

            # Get the observation time span and compute the horizontal
            # position of the SSO
            times = obs.shared[self.times].data

            (source_az, source_el, source_dist, source_diameter) = (
                self._get_source_position(obs, observer, times)
            )

            # Make sure detector data output exists
            dets = obs.select_local_detectors(detectors)

            obs.detdata.ensure(self.det_data, detectors=dets, create_units=u.K)

            det_units = obs.detdata[self.det_data].units

            scale = unit_conversion(u.K, det_units)

            self._observe_source(
                data,
                obs,
                source_az,
                source_el,
                source_dist,
                source_diameter,
                prefix,
                dets,
                scale,
                times,
            )

        if data.comm.group_rank == 0:
            timer.stop()
            log.debug(
                f"{data.comm.group} : Simulated and observed Source in "
                f"{timer.seconds():.1f} seconds"
            )

        return

    def _check_scan_type(self, scan_string):

        log = Logger.get()

        available_scans = [
            "elevation_only",
            "azimuth_only",
            "elevation_only_single",
            "azimuth_only_single",
            "grid",
            "fixed",
        ]

        if scan_string in available_scans:
            return scan_string
        else:
            log.debug("Invalid Scan Type input" "Set scan to elevation_only")

            return "elevation_only"

    @function_timer
    def _get_source_position(self, obs, observer, times):

        log = Logger.get()
        timer = Timer()
        timer.start()

        scan_type = self._check_scan_type(self.source_scan_type)

        az_telescope_range = np.ptp(np.array(obs.shared[self.azimuth])) * u.rad
        el_telescope_range = np.ptp(np.array(obs.shared[self.elevation])) * u.rad

        max_telescope_range = (
            np.amax(np.array([az_telescope_range.value, el_telescope_range.value]))
            * u.rad
        )

        if self.source_azimuth_range == 0:
            az_range = max_telescope_range
        else:
            az_range = self.source_azimuth_range
        if self.source_elevation_range == 0:
            el_range = max_telescope_range
        else:
            el_range = self.source_elevation_range

        if scan_type == "fixed":

            if self.source_azimuth_start.value == 0:
                az_start = np.median(np.array(obs.shared[self.azimuth])) * u.rad
            else:
                az_start = self.source_azimuth_start

            if self.source_elevation_start.value == 0:
                el_start = np.median(np.array(obs.shared[self.elevation])) * u.rad
            else:
                el_start = self.source_elevation_start

        elif scan_type == "azimuth_only" or scan_type == "azimuth_only_single":

            if self.source_azimuth_start.value != 0:
                az_start = self.source_azimuth_start
            else:
                if self.source_azimuth_direction == "increasing":
                    az_start = np.amin(np.array(obs.shared[self.azimuth])) * u.rad
                    az_start -= az_range / 2
                else:
                    az_start = np.amax(np.array(obs.shared[self.azimuth])) * u.rad
                    az_start += az_range / 2

            if self.source_elevation_start.value == 0:
                el_start = np.median(np.array(obs.shared[self.elevation])) * u.rad
            else:
                el_start = self.source_elevation_start

        elif scan_type == "elevation_only" or scan_type == "elevation_only_single":

            if self.source_elevation_start.value != 0:
                el_start = self.source_elevation_start
            else:
                if self.source_elevation_direction == "increasing":
                    el_start = np.amin(np.array(obs.shared[self.elevation])) * u.rad
                    el_start -= el_range / 2
                else:
                    el_start = np.amax(np.array(obs.shared[self.elevation])) * u.rad
                    el_start += el_range / 2

            if self.source_azimuth_start.value == 0:
                az_start = np.median(np.array(obs.shared[self.azimuth])) * u.rad
            else:
                az_start = self.source_azimuth_start

        elif scan_type == "grid_scan":

            if self.source_elevation_start.value != 0:
                el_start = self.source_elevation_start
            else:
                if self.source_elevation_direction == "increasing":
                    el_start = np.amin(np.array(obs.shared[self.elevation])) * u.rad
                    el_start -= el_range / 2
                else:
                    el_start = np.amax(np.array(obs.shared[self.elevation])) * u.rad
                    el_start += el_range / 2

            if self.source_azimuth_direction == "increasing":
                az_start = np.amin(np.array(obs.shared[self.azimuth])) * u.rad
                az_start -= az_range / 2
            else:
                az_start = np.amax(np.array(obs.shared[self.azimuth])) * u.rad
                az_start += az_range / 2

        source_scan_params = {
            "azimuth_starting": az_start,
            "azimuth_range": az_range,
            "azimuth_velocity": self.source_azimuth_velocity,
            "azimuth_acceleration": self.source_azimuth_acceleration,
            "azimuth_direction": self.source_azimuth_direction,
            "elevation_starting": el_start,
            "elevation_range": el_range,
            "elevation_velocity": self.source_elevation_velocity,
            "elevation_acceleration": self.source_elevation_acceleration,
            "elevation_direction": self.source_elevation_direction,
            "elevation_step": self.source_elevation_step,
            "scan_time": np.ptp(times) * u.s,
            "scan_type": scan_type,
        }

        scan = SimulateDroneMovement(source_scan_params)

        _, source_azimuth, source_elevation = scan.create_scan(time_interp=times * u.s)

        if np.any(np.array(self.source_err) >= 1e-4) or self.wind_gusts_amp.value != 0:

            E, N, U = local.hor2enu(
                source_azimuth, source_elevation, self.source_distance
            )
            X, Y, Z = local.enu2ecef(
                E, N, U, observer.lon, observer.lat, observer.elevation, ell="WGS84"
            )

            if np.any(np.array(self.source_err) >= 1e-4):
                X = (
                    X
                    + np.random.normal(0, self.source_err[0], size=(len(times)))
                    * X.unit
                )
                Y = (
                    Y
                    + np.random.normal(0, self.source_err[1], size=(len(times)))
                    * Y.unit
                )
                Z = (
                    Z
                    + np.random.normal(0, self.source_err[2], size=(len(times)))
                    * Z.unit
                )

            if self.wind_gusts_amp.value != 0:

                delta_t = np.amin(np.diff(times))

                samples = int(self.wind_gusts_duration.value / delta_t)

                dx = np.zeros((self.wind_gusts_number, samples))
                dy = np.zeros_like(dx)
                dz = np.zeros_like(dx)

                # Compute random wind direction for any wind gust
                v = np.random.rand(self.wind_gusts_number, 3)
                wind_direction = v / np.linalg.norm(v)

                # Compute the angles using the versor direction
                theta = np.arccos(wind_direction[:, 2])
                phi = np.arctan2(wind_direction[:, 1], wind_direction[:, 0])

                base = np.reshape(
                    np.tile(np.arange(0, samples + 1), self.wind_gusts_number),
                    (self.wind_gusts_number, samples + 1),
                )

                dt = base * delta_t

                wind_amp = self.wind_gusts_amp * self.drone_damp

                dz = wind_amp * wind_direction[:, 2][:, np.newaxis] * dt
                dx = (
                    wind_amp
                    * np.sin(theta[:, np.newaxis])
                    * np.cos(phi[:, np.newaxis])
                    * dt
                )
                dy = (
                    wind_amp
                    * np.sin(theta[:, np.newaxis])
                    * np.sin(phi[:, np.newaxis])
                    * dt
                )

                # Create an array of position returning to the origin
                dx = np.hstack((dx, np.flip(dx, axis=1)))
                dy = np.hstack((dy, np.flip(dy, axis=1)))
                dz = np.hstack((dz, np.flip(dz, axis=1)))

                idx = np.arange(len(X))
                idx_wind = np.ones(self.wind_gusts_number)

                while np.any(np.diff(idx_wind) < 2.5 * samples):
                    idx_wind = np.random.choice(idx, self.wind_gusts_number)

                idxs = (
                    np.hstack(
                        (
                            base,
                            base[:, -1][:, np.newaxis]
                            + np.ones(self.wind_gusts_number, dtype=int)[:, np.newaxis]
                            + base,
                        )
                    )
                    + idx_wind
                ).flatten()

                (good,) = np.where(idxs < len(X))
                valid = np.arange(0, len(good), dtype=int)

                X[idxs[good]] += dx.flatten()[valid]
                Y[idxs[good]] += dy.flatten()[valid]
                Z[idxs[good]] += dz.flatten()[valid]

            X_tel, Y_tel, Z_tel, _, _, _ = local.lonlat2ecef(
                observer.lon, observer.lat, observer.elevation
            )

            E, N, U, _, _, _ = local.ecef2enu(
                X_tel,
                Y_tel,
                Z_tel,
                X,
                Y,
                Z,
                0,
                0,
                0,
                0,
                0,
                0,
                observer.lon,
                observer.lat,
            )

            source_az, source_el, source_dist, _, _, _ = local.enu2hor(E, N, U, 0, 0, 0)

        else:
            source_az = source_azimuth.copy()
            source_el = source_elevation.copy()
            source_dist = self.source_distance.copy()

        size = local._check_quantity(self.source_size, u.m)
        size = (size / source_dist) * u.rad

        obs["source_az"] = source_az
        obs["source_el"] = source_el
        obs["source_distance"] = source_dist

        # Create a shared data object with the source location
        source_coord = np.column_stack(
            [-source_az.to_value(u.degree), source_el.to_value(u.degree)]
        )
        obs.shared.create_column("source", (len(source_az), 2), dtype=np.float64)
        if obs.comm.group_rank == 0:
            obs.shared["source"].set(source_coord)
        else:
            obs.shared["source"].set(None)

        if obs.comm.group_rank == 0:
            timer.stop()
            log.verbose(
                f"{obs.comm.group} : Computed source position in "
                f"{timer.seconds():.1f} seconds"
            )

        return source_az, source_el, source_dist, size

    def _get_source_temp(self):

        power = self.source_power
        fc = self.source_fc
        width = self.source_width
        gain = self.source_gain
        diameter = self.source_size
        amplitude = self.source_amplitude
        noise_bw = self.source_noise_bw
        noise_out = self.source_noise_out

        freq, spec = source_spectrum(
            fc, width, power, gain, amplitude, diameter, noise_bw, noise_out
        )

        temp = utils.s2tcmb(spec, freq)

        return freq, temp

    def _get_drone_temp(self):
        """
        Compute the drone CMB temperature as a grey body
        """

        freqs = np.linspace(20.0, 350.0, 1001, endpoint=True) * 1e9 * u.Hz
        temp = utils.tb2tcmb(self.drone_temp, freqs) * self.drone_emiss

        return freqs, temp

    def _get_beam_map(self, det, source_diameter, ttemp_det):
        """
        Construct a 2-dimensional interpolator for the beam
        """
        # Read in the simulated beam.  We could add operator traits to
        # specify whether to load different beams based on detector,
        # wafer, tube, etc and check that key here.
        if "ALL" in self.beam_props:
            # We have already read the single beam file.
            beam_dic = self.beam_props["ALL"]
        else:
            with h5py.File(self.beam_file, "r") as f_t:
                beam_dic = {}
                beam_dic["data"] = f_t["beam"][:]
                beam_dic["size"] = [
                    [f_t["beam"].attrs["size"], f_t["beam"].attrs["res"]],
                    [f_t["beam"].attrs["npix"], 1],
                ]
                self.beam_props["ALL"] = beam_dic
        description = beam_dic["size"]  # 2d array [[size, res], [n, 1]]
        model = beam_dic["data"]
        res = description[0][1] * u.degree
        beam_solid_angle = np.sum(model) * res**2

        n = int(description[1][0])
        size = description[0][0]
        source_radius_avg = np.average(source_diameter) / 2
        source_solid_angle = np.pi * source_radius_avg**2

        amp = ttemp_det * (
            source_solid_angle.to_value(u.rad**2) / beam_solid_angle.to_value(u.rad**2)
        )
        w = np.radians(size / 2)
        x = np.linspace(-w, w, n)
        y = np.linspace(-w, w, n)
        model *= amp
        beam = RectBivariateSpline(x, y, model)
        r = np.sqrt(w**2 + w**2)
        return beam, r

    @function_timer
    def _observe_source(
        self,
        data,
        obs,
        source_az,
        source_el,
        source_dist,
        source_diameter,
        prefix,
        dets,
        scale,
        times,
    ):
        """
        Observe the Source with each detector in tod
        """
        log = Logger.get()
        timer = Timer()

        source_freq, source_temp = self._get_source_temp()

        # Get a view of the data which contains just this single
        # observation
        obs_data = data.select(obs_uid=obs.uid)

        for det in dets:
            timer.clear()
            timer.start()
            bandpass = obs.telescope.focalplane.bandpass
            signal = obs.detdata[self.det_data][det]

            # Compute detector quaternions and Stokes weights

            self.detector_pointing.apply(obs_data, detectors=[det])
            if self.detector_weights is not None:
                self.detector_weights.apply(obs_data, detectors=[det])

            azel_quat = obs.detdata[self.detector_pointing.quats][det]

            # Convert Az/El quaternion of the detector into angles
            theta, phi, _ = qa.to_iso_angles(azel_quat)

            # Azimuth is measured in the opposite direction
            # than longitude
            az = 2 * np.pi - phi
            el = np.pi / 2 - theta

            # Convolve the planet SED with the detector bandpass
            det_temp = bandpass.convolve(det, source_freq, source_temp)

            beam, radius = self._get_beam_map(det, source_diameter, det_temp)

            # Interpolate the beam map at appropriate locations
            az_diff = (az - source_az.to_value(u.rad) + np.pi) % (2 * np.pi) - np.pi
            x = az_diff * np.cos(el)
            y = el - source_el.to_value(u.rad)
            r = np.sqrt(x**2 + y**2)
            good = r < radius

            sig = np.zeros(len(times))
            sig[good] = beam(x[good], y[good], grid=False)

            if self.source_freq_chopping.value > 0:
                sampling = 1 / np.mean(np.diff(times))
                chop = square(2 * np.pi * sampling * self.source_freq_chopping)
                chop[chop < 1] = 0

                sig *= chop

            if self.drone_temp > 250 * u.K and self.drone_emiss > 0:

                drone_diameter = local._check_quantity(self.drone_size, u.m)
                drone_diameter = (drone_diameter / source_dist) * u.rad

                drone_freq, drone_temp = self._get_drone_temp()
                det_drone_temp = bandpass.convolve(det, drone_freq, drone_temp)

                beam_drone, _ = self._get_beam_map(det, drone_diameter, det_drone_temp)

                sig_drone = np.zeros(len(times))
                sig_drone[good] = beam_drone(x[good], y[good], grid=False)

                sig += sig_drone

            # Stokes weights for observing polarized source
            if self.detector_weights is None:
                weights_I = 1
                weights_Q = 0
                weights_U = 0
            else:
                weights = obs.detdata[self.detector_weights.weights][det]
                weight_mode = self.detector_weights.mode
                if "I" in weight_mode:
                    ind = weight_mode.index("I")
                    weights_I = weights[:, ind].copy()
                else:
                    weights_I = 0
                if "Q" in weight_mode:
                    ind = weight_mode.index("Q")
                    weights_Q = weights[:, ind].copy()
                else:
                    weights_Q = 0
                if "U" in weight_mode:
                    ind = weight_mode.index("U")
                    weights_U = weights[:, ind].copy()
                else:
                    weights_U = 0

            pfrac = self.polarization_fraction
            angle = (
                self.source_pol_angle
                + np.random.normal(
                    0, self.source_pol_angle_error.to(u.deg).value, size=(len(sig))
                )
                * u.deg
            )

            I = sig.copy()
            Q = pfrac * sig * np.cos(2 * angle)
            U = pfrac * sig * np.sin(2 * angle)

            drone_sig = (I * weights_I + Q * weights_Q + U * weights_U).value

            signal += drone_sig

            timer.stop()
            if data.comm.world_rank == 0:
                log.verbose(
                    f"{prefix} : Simulated and observed source in {det} in "
                    f"{timer.seconds():.1f} seconds"
                )

        return

    def _finalize(self, data, **kwargs):
        return

    def _requires(self):
        req = {
            "shared": [
                self.times,
            ],
            "detdata": [
                self.det_data,
                self.quats_azel,
            ],
        }

        if self.weights is not None:
            req["weights"].append(self.weights)

        return req

    def _provides(self):
        return {
            "detdata": [
                self.det_data,
            ],
            "shared": [
                "source",
            ],
        }

    def _accelerators(self):
        return list()



import os
import traceback
import argparse

import datetime as dt
from typing import Optional
from sotodlib.io.imprinter import Imprinter, Books, FAILED
import sotodlib.io.imprinter_utils as utils
import sotodlib.site_pipeline.util as sp_util

import multiprocessing
from concurrent.futures import ProcessPoolExecutor, as_completed


def bind_books_parallel(platform, book_list, n_proc ):
    imprint = Imprinter.for_platform(platform)
    session = imprint.get_session()
    bid_list = [book.bid for book in book_list]
    failed_list = []
    with ProcessPoolExecutor(n_proc) as exe:
        futures = [
            exe.submit(_bookbinding_helper, platform, bid) for bid in bid_list
        ]
        for future in as_completed(futures):
            bid, status, message, _  = future.result()
            imprint.logger.info(f"Just finished book {bid}")
            book = session.query(Books).filter(Books.bid == bid).one()
            book.status = status
            book.message = message
            session.commit()
            if status == FAILED:
                failed_list.append( book.bid)
    return failed_list

def _bookbinding_helper(platform, bid ):
    imprint = Imprinter.for_platform(platform)
    return imprint._run_book_binding(bid)


def main(config: str, n_proc:int=1, alert_webhook: str=''):
    """Make books based on imprinter db
    
    Parameters
    ----------
    config : str
        path to imprinter configuration file
    n_proc : int
        Number of processes
    alert_webhook : str
        Webhook URL to send alerts
    """
    imprinter = Imprinter(
        config, 
        db_args={'connect_args': {'check_same_thread': False}}
    )
    
    # get unbound books
    unbound_books = imprinter.get_unbound_books()
    already_failed_books = imprinter.get_failed_books()
    
    if n_proc>1:
        multiprocessing.set_start_method('spawn')
        parallel_list = [
            book for book in unbound_books if book.type == 'oper'
        ]
        bind_books_parallel(imprinter.daq_node, parallel_list, n_proc=n_proc)

    unbound_books = imprinter.get_unbound_books()
    print(f"Found {len(unbound_books)} unbound books and "
        f"{len(already_failed_books)} failed books")
    for book in unbound_books:
        print(f"Binding book {book.bid}")
        try:
            imprinter.bind_book(book)
        except Exception as e:
            print(f"Error binding book {book.bid}: {e}")
            print(traceback.format_exc())


    print("Retrying failed books")
    failed_books = imprinter.get_failed_books()
    for book in failed_books:
        if book in already_failed_books:
            print(f"Book {book.bid} has already failed twice, not re-trying")
            continue
        print(f"Binding book {book.bid}")
        if 'NoHWPData' in book.message:
            print(
                f"Book {book.bid} does not HWP data reading out, binding "
                    "anyway"
            )
            require_hwp = False
        else:
            require_hwp = True    
        try:         
            utils.set_book_rebind(imprinter, book)
            imprinter.bind_book(book, require_hwp=require_hwp)
        except Exception as e:
            print(f"Error binding book {book.bid}: {e}")
            print(traceback.format_exc())
            # it has failed twice, ideally we want people to look at it now
            # do something here
            alert = sp_util.send_alert(alert_webhook, alertname=book.bid, tag='bookbinder', error=str(e), timestamp=book.start)
            print(alert)


def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    parser.add_argument(
        'config', 
        type=str, 
        help="Path to imprinter configuration file"
    )
    parser.add_argument(
        "--n-proc", type=int, default=1,
        help="The number of processes to run for operations books"
    )
    parser.add_argument(
        "--alert-webhook", type=str, default='',
        help="Webhook address to send error alerts"
    )
    return parser


if __name__ == "__main__":
    parser = get_parser(parser=None)
    args = parser.parse_args()
    main(**vars(args))



import numpy as np
import so3g
from . import pca
import logging

logger = logging.getLogger(__name__)

class Extract:
    """Container for storage of sparse sub-segments of a vector.  This
    ties together a Ranges object and a data array with the same
    length as the total length of the Ranges segments.

    This is useful in cases where a relatively small number of samples
    need to be excised from a signal vector, but stored and perhaps
    restored at a later time.

    In docstrings below a "full data vector" is a 1-d array with
    self.ranges.count elements; the "tracked samples" are a subset
    consisting of self.ranges.mask.sum() elements.

    Attributes:
        ranges: The so3g.proj.Ranges object mapping tracked samples
            into full data vector.
        n_ex: The number of tracked samples.
        data: The data vector containing tracked samples.
    """
    def __init__(self, ranges, init_data=True):
        """Constructor.

        Arguments:
            ranges: a Ranges object; all positive ranges will be
                tracked.
            init_data: array or boolean.  If False, self.data is
                initialized to None.  If True, self.data is
                initialized to zeros.  Otherwise, init_data is stored
                in self.data.  Note this is a reference, not a copy.

        """
        self.ranges = ranges.copy()
        rr = self.ranges.ranges()
        self.n_ex = rr[:, 1].sum() - rr[:, 0].sum()
        if init_data is False or init_data is None:
            self.data = None
        elif init_data is True:
            self.data = np.zeros(self.n_ex)
        else:
            self.data = init_data

    @property
    def dtype(self):
        return self.data.dtype

    @property
    def shape(self):
        return (self.ranges.count,)

    def offset_iter(self):
        """Returns an iterator for use in expanding / collapsing extracts.
        Each iteration returns a tuple (ex_lo, ex_hi, full_lo,
        full_hi) where ex_lo:ex_hi is a range in extracted data
        vector, and full_lo:full_hi is a range in the full data
        vector.

        """
        offset = 0
        for lo, hi in self.ranges.ranges():
            yield offset, offset + hi - lo, lo, hi
            offset += hi - lo

    def expand(self, fill_value=0):
        """Expands the extract into a full-length data array, filling missing
        bits with fill_value."""
        output = np.zeros(self.ranges.count, self.data.dtype) + fill_value
        for elo, ehi, lo, hi in self.offset_iter():
            output[lo:hi] = self.data[elo:ehi]
        return output

    def swap(self, signal):
        """Swaps the current extract with the tracked samples of full data
        vector signal."""
        to_save = np.empty(self.n_ex, self.data.dtype)
        for elo, ehi, lo, hi in self.offset_iter():
            to_save[elo:ehi] = signal[lo:hi]
            signal[lo:hi] = self.data[elo:ehi]
        self.data[:] = to_save

    def patch(self, signal):
        """Copies the extract into the signal vector.  Untracked samples are
        not modified.

        """
        for elo, ehi, lo, hi in self.offset_iter():
            signal[lo:hi] = self.data[elo:ehi]
        return signal

    def accumulate(self, signal, scale):
        """Adds the extract, scaled by factor scale, into the full length
        signal vector.  Untracked samples are not modified.

        """
        for elo, ehi, lo, hi in self.offset_iter():
            signal[lo:hi] += self.data[elo:ehi] * scale
        return signal


class ExtractMatrix:
    """This class is a simple container for a list (of length n_dets) of
    Extract objects of equal length (n_samps), to provide abstract
    access to data with shape (n_dets, n_samps).  Each child Extract
    can be accessed with [item] indexing.  Most methods have (tod,
    signal=None) signature, and are simple loops that call the method
    of the same name on each child Extract object.

    """
    def __init__(self, items=None):
        """The complete list of Extract objects should be passed in."""
        if items is None:
            items = []
        self.items = items

    def __repr__(self):
        """This repr shows the shape of the full array, along with the
        fraction (as a percentage) of the full space that is actually
        tracked.

        """
        frac = (sum([e.n_ex for e in self.items]) /
                (self.shape[0] * self.shape[1]))
        return 'ExtractMatrix(' + ','.join(map(str, self.shape)) + \
            '@%.1f%%)' % (frac*100)

    def __len__(self):
        return self.shape[0]

    def copy(self):
        return ExtractMatrix([x.copy() for x in self.items])

    @property
    def shape(self):
        if len(self.items) == 0:
            return (0,)
        return (len(self.items),) + self.items[0].shape

    def __getitem__(self, index):
        return self.items[index]

    def expand(self, fill_value=0):
        """Expands the extract into a full-size data array, filling missing
        bits with fill_value."""
        dtype = self.items[0].dtype
        signal = np.zeros(self.shape, dtype) + fill_value
        for o, i in zip(signal, self.items):
            i.patch(o)
        return signal

    def swap(self, tod, signal=None):
        """Swaps the current extract with the tracked samples of full data
        vector signal."""
        if signal is None:
            signal = tod.signal
        for o, i in zip(signal, self.items):
            i.swap(o)
        return signal

    def patch(self, tod, signal=None):
        """Copies the extract into the signal vector.  Untracked samples are
        not modified.

        """
        if signal is None:
            signal = tod.signal
        for o, i in zip(signal, self.items):
            i.patch(o)
        return signal

    def accumulate(self, tod, signal, scale):
        """Adds the extract, scaled by factor scale, into the full length
        signal vector.  Untracked samples are not modified.

        """
        if signal is None:
            signal = tod.signal
        for o, i in zip(signal, self.items):
            i.accumulate(o, scale)
        return signal


def get_gap_fill_single(data, flags, nbuf=10, order=1, swap=False):
    """Computes samples to fill the gaps in data identified by flags.
    Each flagged segment is modeled with a polynomial of the order
    specified, based on up to nbuf points on each side of the segment.

    Arguments:
        data: 1d vector of samples.
        flags: Ranges object consistent with data size.
        nbuf: Maximum number of samples on each side of a flagged span
            to use for model fit.  If the unflagged span between two
            flagged spans is shorter than nbuf, then it is still used
            to anchor one end of the fit.
        order: Maximum order of polynomial model used in
            interpolation.  The actual order may be smaller if
            insufficient data are available to constrain the
            coefficients.  Care should be taken when going higher than
            linear (order=1); this will tend to be unstable unless the
            gaps are much smaller than the anchors used to constrain
            the poly on each end.
        swap: If False, do not modify the input data vector.  If
            True, patch the data with the model.

    Returns:
        An Extract object containing the modeled data.  If swap is
        True, then the input data vector is patched with the model and
        the returned object, instead, contains the samples from data
        that were changed.

    """
    rsegs = (flags.copy().buffer(nbuf) * ~flags)
    rseg_ranges = rsegs.ranges()
    A = np.zeros((order+1, order+1))
    b = np.zeros(order+1)
    t0, y0 = 0, 0
    model = None
    model_i = -1  # Set to trigger update.
    sig_ex = Extract(flags)

    for elo, ehi, lo, hi in sig_ex.offset_iter():
        while (model_i + 1 < len(rseg_ranges) and
               lo > rseg_ranges[model_i+1][0]):
            model = None
            model_i += 1
        if model is None:
            t0, y0 = lo, data[lo-1]
            b[:] = 0
            A[:] = 0
            contrib_count = 0
            for f in [model_i, model_i+1]:
                if f < 0 or f >= len(rseg_ranges):
                    continue
                _lo, _hi = rseg_ranges[f]
                _t = np.arange(_lo, _hi) - t0
                for _j in range(order+1):
                    b[_j] += np.dot(_t**_j, data[_lo:_hi] - y0)
                    for _k in range(_j, order+1):
                        A[_j, _k] += (_t**(_j+_k)).sum()
                        A[_k, _j] = A[_j, _k]
                contrib_count += _hi - _lo
            if contrib_count == 0:
                y0 = 0.
                model = [0.]
            else:
                # Only fit as many terms as you plausibly constrain --
                # 10 data points per term.
                n_keep = 1 + max(0, min(order, contrib_count // 10))
                model = np.linalg.solve(A[:n_keep,:n_keep], b[:n_keep])[::-1]

        t = np.arange(lo, hi) - t0
        sig_ex.data[elo:ehi] = np.polyval(model, t) + y0
    if swap:
        sig_ex.swap(data)
    return sig_ex


def get_gap_fill(tod, nbuf=10, order=1, swap=False, signal=None, flags=None,
                 _method='fast'):
    """See get_gap_fill_single for meaning of arguments not described here.

    Arguments:
        tod: AxisManager with (dets, samps) axes.
        signal: signal to pass to get_gap_fill_single as data
            argument; defaults to tod.signal
        flags: flags to pass to get_gap_fill_single; defaults to
            tod.flags.

    Returns:
        The ExtractMatrix object with per-detector Extracts from
        get_gap_fill_single.

    """
    if signal is None:
        signal = tod.signal
    if flags is None:
        flags = tod.flags
    if _method is None:
        _method = 'fast' if hasattr(so3g, 'get_gap_fill_poly') else 'slow'

    if _method == 'fast':
        sample_counts = [np.dot(r.ranges(), [-1, 1]).sum()
                         for r in flags]
        dest = np.empty(sum(sample_counts), dtype='float32')
        so3g.get_gap_fill_poly(flags, signal, nbuf, order, swap, dest)
        sample_idx = np.cumsum([0] + sample_counts)
        return ExtractMatrix([Extract(r, dest[i:j])
                              for r, i, j in zip(flags, sample_idx[:-1], sample_idx[1:])])
    else:
        return ExtractMatrix([get_gap_fill_single(d, f, order=order, nbuf=nbuf, swap=swap)
                              for d, f in zip(signal, flags)])


def get_gap_model_single(weights, modes, flags):
    """Computes samples to fill gaps in data identified by flags, based on
    weights and modes (such as would be contained in a PCA model).

    Arguments:
        weights: 1-d array of weights (n_mode) to apply to each mode.
        modes: 2-d array of modes (n_mode, n_samps)
        flags: so3g.proj.Ranges object with count == n_samps.

    Returns:
        An Extract object containing the modeled data.

    """
    sig_ex = Extract(flags)
    for w, m in zip(weights, modes):
        for elo, ehi, lo, hi in sig_ex.offset_iter():
            sig_ex.data[elo:ehi] += w * m[lo:hi]
    return sig_ex


def get_gap_model(tod, model, flags=None, weights=None, modes=None):
    """Calls get_gap_model_single on each detector and its corresponding
    model weights.

    Arguments:
        tod: AxisManager with (dets, samps) axes.
        model: AxisManager with (dets, eigen, samps) axes.
        flags: flags to pass to get_gap_model_single; defaults to
            tod.flags.
        weights: array of mode couplings (dets, eigen); defaults to
            model.weights.
        modes: array with time-dependent modes (eigen, samps);
            defaults to model.modes.

    Returns:
        The ExtractMatrix object with per-detector models from
        get_gap_model_single.

    """
    if flags is None:
        flags = tod.flags
    if weights is None:
        weights = model.weights
    if modes is None:
        modes = model.modes
    return ExtractMatrix([get_gap_model_single(w, modes, f)
                          for w, f in zip(weights, flags)])

def get_contaminated_ranges(good_flags, bad_flags):
    """Determine what intervals in good_flags are contaminated (overlap
    with) intervals in bad_flags.  Note this isn't as simple as
    good_flags * bad_flags, because any contiguous region in
    good_flags is considered contaminated if even one sample of it is
    touched by bad_flags.

    Args:
      good_flags (RangesMatrix): The flags to check for contamination.
        Must have shape (dets, samps).
      bad_flags (RangesMatrix): The flags marking bad data, which
        should be considered as a contaminant to good_flags.  Same
        shape as good_flags.

    Returns:
      RangesMatrix with same shape as inputs, indicating the intervals
      from good_flags that overlap at all with some interval of
      bad_flags.

    Example:
      Move contaminated intervals into bad_flags::

        contam = get_contaminated_ranges(source_flags, glitch_flags)
        source_flags *= ~contam
        glitch_flags += contam

    """
    contam = good_flags.zeros_like()
    for r0, r1, rs in zip(good_flags.ranges, bad_flags.ranges,
                          contam.ranges):
        overlap = (r0 * r1).ranges()
        if len(overlap) == 0:
            continue
        # Any interval in r0 that overlaps must be moved
        for i0, i1 in r0.ranges():
            if np.any((i0 <= overlap[:,0]) * (overlap[:,0] < i1)):
                rs.add_interval(int(i0), int(i1))
    return contam


def fill_glitches(aman, nbuf=10, use_pca=False, modes=3, signal=None,
                  glitch_flags=None, in_place=True, wrap=None):
    """
    This function fills pre-computed glitches provided by the caller in
    time-ordered data using either a polynomial (default) or PCA-based
    approach. Wraps the other functions in the ``tod_ops.gapfill`` module.

    Args
    -----
    aman : AxisManager
        AxisManager to fill glitches in
    nbuf : int
        Number of buffer samples to use in polynomial gap filling.
    use_pca : bool
        Whether or not to fill glitches using pca model. Default is False
    modes : int
        Number of modes in the pca to use if pca=True. Default is 3.
    signal : ndarray or None
        Array of data to fill glitches in. If None then uses ``aman.signal``.
        Default is None.
    glitch_flags : str or RangesMatrix or None
        RangesMatrix containing flags to use for gap filling. If provided by
        a string, ``aman.flags.get(flags)`` is used for the flags. If None then
        uses ``aman.flags.glitches``.
    in_place : bool
        If False it makes a copy of signal before gap filling and returns
        the copy.
    wrap : str or None
        If not None, wrap the gap filled data into tod with this name.

    Returns
    -------
    signal : ndarray
        Returns ndarray with gaps filled from input signal.
    """
    # Process Args
    if signal is None:
        if in_place:
            sig = aman.signal
        else:
            sig = np.copy(aman.signal)
    else:
        if in_place:
            sig = signal
        else:
            sig = np.copy(signal)

    if glitch_flags is None:
        glitch_flags = aman.flags.glitches
    if isinstance(glitch_flags, str):
        glitch_flags = aman.flags.get(glitch_flags)

    # Polyfill
    gaps = get_gap_fill(aman, nbuf=nbuf, flags=glitch_flags,
                        signal=np.float32(sig))
    sig = gaps.swap(aman, signal=sig)
    
    #PCA Fill
    if use_pca:
        if modes > aman.dets.count:
            logger.warning(f'modes = {modes} > number of detectors = ' +
                           f'{aman.dets.count}, setting modes = number of ' +
                           'detectors')
            modes = aman.dets.count
        # PCA fill
        mod = pca.get_pca_model(tod=aman, n_modes=modes,
                                signal=sig)
        gfill = get_gap_model(tod=aman, model=mod, flags=glitch_flags)
        sig = gfill.swap(aman, signal=sig)
    
    # Wrap and Return
    if wrap is not None:
        if wrap in aman._assignments:
            aman.move(wrap, None)
        aman.wrap(wrap, sig, [(0, 'dets'), (1, 'samps')])
        
    return sig



# Copyright (c) 2023-2023 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Interval operations.
"""

import numpy as np
from astropy import units as u
import toast
import toast.ops

from .. import ops as so_ops
from .job import workflow_timer


def setup_az_intervals(operators):
    """Add commandline args and operators for building Azimuth intervals.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(
        toast.ops.AzimuthIntervals(
            name="az_intervals",
            cut_short=True,
            short_limit=5.0 * u.second,
            enabled=False,
        )
    )


@workflow_timer
def create_az_intervals(job, otherargs, runargs, data):
    """Pass through pointing and create Az intervals

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.az_intervals.enabled:
        job_ops.az_intervals.apply(data)



#!/usr/bin/env python3

# Copyright (c) 2024-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""
Helper script for batching observations.

The results are printed to stdout to enable use of this
script inside a slurm script.

"""

import argparse
import os
import re
import time


def load_obs_file(path):
    """Load the observation file into a list.

    Args:
        path (str):  Path to the file

    Returns:
        (list):  The list of observation IDs.

    """
    olist = list()
    with open(path, "r") as f:
        for line in f:
            if re.match(r"^#.*", line) is None:
                olist.append(line.strip())
    return olist


def get_state_file(out_root, obs, name="state"):
    """Construct the path to the state file.

    Args:
        out_root (str):  The top-level output directory
        obs (str):  The observation ID

    Returns:
        (str):  The state file path.

    """
    obs_dir = os.path.join(out_root, obs)
    state_file = os.path.join(obs_dir, name)
    return state_file


def get_obs_state(out_root, obs):
    """Find the current processing state of an observation.
    """
    state_file = get_state_file(out_root, obs)
    if os.path.isfile(state_file):
        with open(state_file, "r") as f:
            state = f.readline()
        return state.rstrip()
    return None


def clear_obs_state(out_root, obs):
    """Clear the current processing state of an observation.
    """
    obs_dir = os.path.join(out_root, obs)
    if not os.path.isdir(obs_dir):
        return
    state_file = get_state_file(out_root, obs)
    if os.path.isfile(state_file):
        os.remove(state_file)


def set_obs_state(out_root, obs, state):
    """Set the current processing state of an observation.
    """
    obs_dir = os.path.join(out_root, obs)
    if not os.path.exists(obs_dir):
        os.makedirs(obs_dir)
    state_file = get_state_file(out_root, obs)
    temp_state = get_state_file(out_root, obs, name="state_temp")
    # Although the temp file only exists for a fraction of a second,
    # check if it was somehow left behind by a previous process.
    if os.path.isfile(temp_state):
        msg = "Temporary state file {temp_state} exists."
        msg += "  This means that a previous batch_control job was "
        msg += "killed while setting the state.  You should cleanup "
        msg += "manually."
        raise RuntimeError(msg)
    with open(temp_state, "w") as f:
        f.write(f"{state}\n")
    os.rename(temp_state, state_file)


def find_obs(all_obs, n_job_obs, out_root, ignore_running=False, timeout_hours=24):
    """Get the list of obs to run for a given job.

    Args:
        all_obs (list):  The list of all observation IDs to consider.
        n_job_obs (int):  The number of observations to select.
        out_root (str):  The output root directory.
        ignore_running (bool):  If True, consider "running" jobs to be eligible
            for re-running.
        timeout_hours (float):  Number of hours after which a "running" job is
            considered dead, regardless of ignore_running option.

    Returns:
        None

    """
    timeout_sec = timeout_hours * 3600
    selected = list()
    for obs in all_obs:
        state = get_obs_state(out_root, obs)
        if state is not None:
            if state == "done":
                continue
            if state == "running" and not ignore_running:
                # Check how long this job has been "running"
                cur_time = time.time()
                state_file = get_state_file(out_root, obs)
                state_time = os.path.getmtime(state_file)
                elapsed = cur_time - state_time
                if elapsed < timeout_sec:
                    # The job has not timed out yet
                    continue
        # We are going to consider this obs
        selected.append(obs)
        if len(selected) >= n_job_obs:
            break
    return selected


def main():
    parser = argparse.ArgumentParser(description="Get observation IDs for a job")
    parser.add_argument(
        "--out_root",
        required=True,
        default=None,
        help="The output root directory",
    )
    parser.add_argument(
        "--observations",
        required=False,
        default=None,
        help="File of observation IDs",
    )
    parser.add_argument(
        "--get_batch",
        required=False,
        type=int,
        default=None,
        help="Get the next batch of observations",
    )
    parser.add_argument(
        "--batch_list",
        required=False,
        action="store_true",
        default=False,
        help="Print the batch as a comma-separated list instead of one per line",
    )
    parser.add_argument(
        "--get_state",
        required=False,
        type=str,
        default=None,
        help="Get the state of this observation",
    )
    parser.add_argument(
        "--set_state_done",
        required=False,
        type=str,
        default=None,
        help="Set the state of this observation to 'done'",
    )
    parser.add_argument(
        "--set_state_running",
        required=False,
        type=str,
        default=None,
        help="Set the state of this observation to 'running'",
    )
    parser.add_argument(
        "--clear_state",
        required=False,
        type=str,
        default=None,
        help="Clear the state of this observation",
    )
    parser.add_argument(
        "--cleanup",
        required=False,
        action="store_true",
        default=False,
        help="Remove the running state from all observations",
    )

    args = parser.parse_args()

    if args.get_batch is not None:
        # We are getting the next batch of observations
        all_obs = load_obs_file(args.observations)
        batch_obs = find_obs(all_obs, args.get_batch, args.out_root)
        if args.batch_list:
            batch_str = ",".join(batch_obs)
            print(f"{batch_str}", flush=True)
        else:
            for obs in batch_obs:
                print(f"{obs}", flush=True)
    elif args.get_state is not None:
        state = get_obs_state(args.out_root, args.get_state)
        print(f"{state}", flush=True)
    elif args.clear_state is not None:
        clear_obs_state(args.out_root, args.clear_state)
    elif args.set_state_done is not None:
        set_obs_state(args.out_root, args.set_state_done, "done")
    elif args.set_state_running is not None:
        set_obs_state(args.out_root, args.set_state_running, "running")
    elif args.cleanup:
        all_obs = load_obs_file(args.observations)
        for obs in all_obs:
            state = get_obs_state(args.out_root, obs)
            if state is not None and state == "running":
                clear_obs_state(args.out_root, obs)


if __name__ == "__main__":
    main()



# Copyright (c) 2025-2025 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import numpy as np

import traitlets

import toast
from toast.timing import function_timer, Timer
from toast.traits import (
    trait_docs,
    Int,
    Unicode,
    Tuple,
    Bool,
)
from toast.ops.operator import Operator
from toast.utils import Logger
from toast.observation import default_values as defaults


@trait_docs
class DetBiasCuts(Operator):
    """Use IV and bias step data to cut poorly biased detectors."""

    # Class traits

    API = Int(0, help="Internal interface version for this operator")

    detcal_prefix = Unicode(
        "det_cal:",
        help="The metadata prefix for detector calibration",
    )

    rfrac_range = Tuple(
        (0.1, 0.7),
        help="The (lower, upper) bounds for rfrac selection",
    )

    psat_range = Tuple(
        (),
        help="The (lower, upper) bounds for P_SAT from IV analysis",
    )

    rn_range = Tuple(
        (),
        help="The (lower, upper) range of r_n for det selection",
    )

    si_nan = Bool(False, help="If True, flag dets where s_i is NaN")

    times = Unicode(defaults.times, help="Observation shared key for timestamps")

    det_data = Unicode(
        defaults.det_data,
        help="Observation detdata key for signal",
    )

    det_mask = Int(
        defaults.det_mask_invalid,
        help="Bit mask value for per-detector flagging",
    )

    shared_flags = Unicode(
        defaults.shared_flags,
        allow_none=True,
        help="Observation shared key for common flags",
    )

    det_flags = Unicode(
        defaults.det_flags,
        allow_none=True,
        help="Observation detdata key for detector flags",
    )

    bias_mask = Int(
        defaults.det_mask_invalid,
        help="Bit mask to apply to detectors that are poorly biased",
    )

    @traitlets.validate("det_mask")
    def _check_det_mask(self, proposal):
        check = proposal["value"]
        if check < 0:
            raise traitlets.TraitError("Det mask should be a positive integer")
        return check

    @traitlets.validate("bias_mask")
    def _check_bias_mask(self, proposal):
        check = proposal["value"]
        if check < 0:
            raise traitlets.TraitError("Bias mask should be a positive integer")
        return check

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()

        bg_key = f"{self.detcal_prefix}bg"
        r_tes_key = f"{self.detcal_prefix}r_tes"
        r_frac_key = f"{self.detcal_prefix}r_frac"
        p_sat_key = f"{self.detcal_prefix}p_sat"
        r_n_key = f"{self.detcal_prefix}r_n"
        s_i_key = f"{self.detcal_prefix}s_i"

        # Each process updates the detector flags for their local dets.
        for ob in data.obs:
            bias_flags = dict()
            fp = ob.telescope.focalplane
            cnames = set(fp.detector_data.colnames)
            for det in ob.select_local_detectors(detectors, flagmask=self.det_mask):
                fp_det = fp[det]
                if bg_key not in cnames:
                    msg = f"{ob.name}:{det} no focalplane key '{bg_key}', cutting"
                    log.debug(msg)
                    bias_flags[det] = self.bias_mask
                    continue
                if fp_det[bg_key] < 0:
                    msg = f"{ob.name}:{det} {bg_key}={fp_det[bg_key]} < 0."
                    msg += " Cutting."
                    log.debug(msg)
                    bias_flags[det] = self.bias_mask
                    continue
                if r_tes_key not in cnames:
                    msg = f"{ob.name}:{det} no focalplane key '{r_tes_key}', cutting"
                    log.debug(msg)
                    bias_flags[det] = self.bias_mask
                    continue
                if fp_det[r_tes_key] <= 0:
                    msg = f"{ob.name}:{det} {r_tes_key}={fp_det[r_tes_key]} <= 0."
                    msg += " Cutting."
                    log.debug(msg)
                    bias_flags[det] = self.bias_mask
                    continue
                if r_frac_key not in cnames:
                    msg = f"{ob.name}:{det} no focalplane key '{r_frac_key}', cutting"
                    log.debug(msg)
                    bias_flags[det] = self.bias_mask
                    continue
                if (
                    fp_det[r_frac_key] < self.rfrac_range[0]
                    or fp_det[r_frac_key] > self.rfrac_range[1]
                ):
                    msg = f"{ob.name}:{det} {r_frac_key}={fp_det[r_frac_key]} outside"
                    msg += f" range {self.rfrac_range}. Cutting."
                    log.debug(msg)
                    bias_flags[det] = self.bias_mask
                    continue
                if len(self.psat_range) != 0:
                    if len(self.psat_range) != 2:
                        msg = "psat_range should be a 2-tuple with the low, high values"
                        raise RuntimeError(msg)
                    if p_sat_key not in cnames:
                        msg = f"{ob.name}:{det} no focalplane key '{p_sat_key}',"
                        msg += " but PSAT range specified. Cutting."
                        log.debug(msg)
                        bias_flags[det] = self.bias_mask
                        continue
                    if (
                        fp_det[p_sat_key] * 1e12 < self.psat_range[0]
                        or fp_det[p_sat_key] * 1e12 > self.psat_range[1]
                    ):
                        msg = f"{ob.name}:{det} {p_sat_key}={fp_det[p_sat_key]} outside"
                        msg += f" range {self.psat_range}. Cutting."
                        log.debug(msg)
                        bias_flags[det] = self.bias_mask
                        continue
                if len(self.rn_range) != 0:
                    if len(self.rn_range) != 2:
                        msg = "rn_range should be a 2-tuple with the low, high values"
                        raise RuntimeError(msg)
                    if r_n_key not in cnames:
                        msg = f"{ob.name}:{det} no focalplane key '{r_n_key}',"
                        msg += " but r_n range specified. Cutting."
                        log.debug(msg)
                        bias_flags[det] = self.bias_mask
                        continue
                    if (
                        fp_det[r_n_key] < self.rn_range[0]
                        or fp_det[r_n_key] > self.rn_range[1]
                    ):
                        msg = f"{ob.name}:{det} {r_n_key}={fp_det[r_n_key]} outside"
                        msg += f" range {self.rn_range}. Cutting."
                        log.debug(msg)
                        bias_flags[det] = self.bias_mask
                        continue
                if self.si_nan:
                    if s_i_key not in cnames:
                        msg = f"{ob.name}:{det} no focalplane key '{s_i_key}',"
                        msg += " but s_i NaN check is True. Cutting."
                        log.debug(msg)
                        bias_flags[det] = self.bias_mask
                        continue
                    if np.isnan(fp_det[s_i_key]):
                        msg = f"{ob.name}:{det} {s_i_key} is NaN. Cutting."
                        log.debug(msg)
                        bias_flags[det] = self.bias_mask
                        continue
            ob.update_local_detector_flags(bias_flags)

    def _finalize(self, data, **kwargs):
        return

    def _requires(self):
        req = {
            "meta": list(),
            "shared": [self.times],
            "detdata": [self.det_data],
            "intervals": list(),
        }
        if self.det_flags is not None:
            req["detdata"].append(self.det_flags)
        return req

    def _provides(self):
        return dict()



import unittest
from sotodlib import coords
from pixell import enmap, tilemap
from sotodlib.coords import healpix_utils as hp_utils
from ._helpers import quick_tod
import numpy as np
try:
    import healpy as hp
except:
    hp = False

class PmatTest(unittest.TestCase):
    """ Test coords.pmat.P.

    Check that for_tod, to_map, to_weights, to_inverse_weights, remove_weights and from_map
    all run without crashing, that maps computed with different settings match, and that
    from_map recovers a constant input signal.
    """

    def test_pmat_rectpix(self):
        obs = quick_tod(10, 10000)
        obs.signal[:,:] = 1
        shape, wcs = enmap.fullsky_geometry(res=0.5*coords.DEG)
        comps = 'T'
        out = run_test(obs, (shape, wcs), comps, None, False, False) # Basic
        _ = run_test(obs, None, comps, wcs, False, False) # Use wcs_kernel
        out2 = run_test(obs, tilemap.geometry(shape, wcs, tile_shape=(100, 100)), comps, None, False, True) # Tiled
        assert np.array_equal(out, out2)

    def test_pmat_healpix(self):
        obs = quick_tod(10, 10000)
        obs.signal[:,:] = 1
        nside = 64
        comps = 'T'
        out = run_test(obs, hp_utils.get_geometry(nside), comps, None, True, False) # Basic un-tiled
        out2 = run_test(obs, hp_utils.get_geometry(nside, 4), comps, None, True, True) # Explicit tiling
        assert np.array_equal(out, out2)
        out2 = run_test(obs, hp_utils.get_geometry(nside, 'auto'), comps, None, True, True) # Auto tiling
        assert np.array_equal(out, out2)
        out2 = run_test(obs, None, comps, f'nside={nside}', True, False) # Basic un-tiled with wcs_kernel
        assert np.array_equal(out, out2)
        out2 = run_test(obs, hp_utils.get_geometry(nside, None, 'RING'), comps, None, True, False) # Un-tiled ring
        if hp:
            out2 = hp.reorder(out2, r2n=True)
            assert np.array_equal(out, out2)

def run_test(obs, geom, comps, wcs_kernel, is_healpix, is_tiled):
    pmat = coords.pmat.P.for_tod(obs, comps=comps, geom=geom, wcs_kernel=wcs_kernel)
    assert pmat.tiled == is_tiled # Check tiled flag works

    imap = pmat.to_map(obs)
    ncomp = len(comps)
    zeros = pmat.zeros((ncomp,ncomp))
    weights = pmat.to_weights(obs, dest=zeros)
    iweights = pmat.to_inverse_weights(weights)
    remove_weights = pmat.remove_weights(tod=obs)
    tod = pmat.from_map(remove_weights)
    TOL = 1e-9
    assert np.all(np.abs(tod-obs.signal) < TOL)

    # Confirm we can do map-space ops without a pointing op first
    pmat = coords.pmat.P.for_tod(obs, comps=comps, geom=geom, wcs_kernel=wcs_kernel)
    _ = pmat.to_inverse_weights(weights)
    pmat = coords.pmat.P.for_tod(obs, comps=comps, geom=geom, wcs_kernel=wcs_kernel)
    _ = pmat.remove_weights(weights)

    # Confirm from_map works on uninitialized pmat
    pmat = coords.pmat.P.for_tod(obs, comps=comps, geom=geom, wcs_kernel=wcs_kernel)
    tod2 = pmat.from_map(remove_weights)
    assert np.all(np.abs(tod - tod2) < TOL)

    # And also zeros.
    pmat = coords.pmat.P.for_tod(obs, comps=comps, geom=geom, wcs_kernel=wcs_kernel)
    pmat.zeros()

    if is_tiled:
        if is_healpix:
            remove_weights = hp_utils.tiled_to_full(remove_weights)
        else:
            remove_weights = tilemap.to_enmap(remove_weights)
    return remove_weights



# Copyright (c) 2018-2023 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Focalplane simulation tools.
"""

import re
from collections import OrderedDict

import astropy.units as u
import numpy as np
from toast.instrument_coords import xieta_to_quat, quat_to_xieta
from toast.instrument_sim import (
    hex_nring,
    hex_xieta_row_col,
    hex_layout,
    rhomb_dim,
    rhomb_xieta_row_col,
    rhombus_hex_layout,
)
import toast.qarray as qa
import toast.utils

from sotodlib.io.metadata import read_dataset
import sotodlib.core.metadata.loader as loader


def sim_wafer_detectors(
    hw,
    wafer_slot,
    platescale,
    fwhm,
    band=None,
    center=None,
):
    """Generate detector properties for a wafer.

    Given a Hardware configuration, generate all detector properties for
    the specified wafer and optionally only the specified band.

    By convention, polarization angle quantities are stored in degrees and
    fwhm is stored in arcminutes.  These units are stripped to ease serialization
    to TOML and JSON.  The units are restored when constructing focalplane tables
    for TOAST.

    Args:
        hw (Hardware): The hardware properties.
        wafer_slot (str): The wafer slot name.
        platescale (float): The plate scale in degrees / mm.
        fwhm (dict): Dictionary of nominal FWHM values in arcminutes for
            each band.
        band (str, optional): Optionally only use this band.
        center (array, optional): The quaternion offset of the center.

    Returns:
        (OrderedDict): The properties of all selected detectors.

    """
    zaxis = np.array([0, 0, 1], dtype=np.float64)

    # The properties of this wafer
    wprops = hw.data["wafer_slots"][wafer_slot]

    # The readout card and its properties
    card_slot = wprops["card_slot"]
    cardprops = hw.data["card_slots"][card_slot]

    # The bands
    bands = wprops["bands"]
    if band is not None:
        if band in bands:
            bands = [band]
        else:
            raise RuntimeError(
                "band '{}' not valid for wafer '{}'".format(band, wafer_slot)
            )

    # Lay out the pixel locations depending on the wafer type.  Also
    # compute the polarization orientation rotation, as well as the A/B
    # handedness for the Sinuous detectors.

    npix = wprops["npixel"]
    pixsep = platescale * wprops["pixsize"] * u.degree
    handed = None
    kill = []
    if wprops["packing"] == "F":
        # Feedhorn (NIST style)
        # The "gap" is the **additional** space beyond the normal pixel
        # separation.  For S.O., we always have the gap equal to nominal
        # pixel spacing.
        gap = 0.0 * u.degree
        nrhombus = npix // 3

        # This dim is also the number of pixels along the short axis.
        dim = rhomb_dim(nrhombus)

        # This is the center-center distance along the short axis
        width = (dim - 1) * pixsep

        # The orientation within each rhombus alternates between zero and 45
        # degrees.  However there is an offset.  We choose this arbitrarily
        # for the nominal rhombus position, and then the rotation of the
        # other 2 rhombi will naturally modulate this.
        pol_A = np.zeros(nrhombus, dtype=np.float64)
        pol_B = np.zeros(nrhombus, dtype=np.float64)
        poloff = 22.5
        for p in range(nrhombus):
            # get the row / col of the pixel
            row, col = rhomb_xieta_row_col(nrhombus, p)
            if np.mod(row, 2) == 0:
                pol_A[p] = 0.0 + poloff
            else:
                pol_A[p] = 45.0 + poloff
            pol_B[p] = 90.0 + pol_A[p]
        pol_A = u.Quantity(pol_A, u.degree)
        pol_B = u.Quantity(pol_B, u.degree)

        layout_A = rhombus_hex_layout(
            nrhombus, width, "", "", gap=gap, pol=pol_A
        )
        layout_B = rhombus_hex_layout(
            nrhombus, width, "", "", gap=gap, pol=pol_B
        )

        # We are going to remove 2 pixels for mechanical reasons
        kf = dim * (dim - 1) // 2
        kill = [(dim * dim + kf), (dim * dim + kf) + dim - 2]

        # The orientation of the wafer with respect to the focalplane
        wafer_rot_rad = 0.0
    elif wprops["packing"] == "S":
        # Sinuous (Berkeley style)
        # This is the center-center distance along the vertex-vertex axis
        width = (2 * (hex_nring(npix) - 1)) * pixsep

        # The orientation of the wafer with respect to the focalplane.  This
        # is 30 degrees.
        wafer_rot_rad = np.pi / 6
        hex_cent = qa.from_axisangle(zaxis, wafer_rot_rad)

        # The sinuous handedness is chosen so that A/B pairs of pixels have the
        # same nominal orientation but trail each other along the
        # vertex-vertex axis of the hexagon.  The polarization orientation
        # changes every other column

        if npix == 37:
            pol_A = np.array(
                [
                    45.0,
                    0.0,
                    45.0,
                    45.0,
                    45.0,
                    45.0,
                    0.0,
                    0.0,
                    0.0,
                    45.0,
                    45.0,
                    0.0,
                    0.0,
                    0.0,
                    45.0,
                    45.0,
                    0.0,
                    0.0,
                    0.0,
                    45.0,
                    0.0,
                    0.0,
                    45.0,
                    45.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    45.0,
                    45.0,
                    0.0,
                    0.0,
                    45.0,
                    45.0,
                    45.0,
                ]
            )
            handed = [
                "R",
                "L",
                "R",
                "L",
                "L",
                "R",
                "L",
                "R",
                "L",
                "R",
                "L",
                "R",
                "R",
                "R",
                "L",
                "R",
                "L",
                "R",
                "R",
                "L",
                "R",
                "L",
                "R",
                "L",
                "R",
                "L",
                "L",
                "L",
                "L",
                "R",
                "L",
                "R",
                "L",
                "R",
                "L",
                "L",
                "L",
            ]
            pol_B = 90.0 + pol_A
        else:
            handed = list()
            pol_A = np.zeros(npix, dtype=np.float64)
            pol_B = np.zeros(npix, dtype=np.float64)
            for p in range(npix):
                row, col = hex_xieta_row_col(npix, p)
                if np.mod(col, 2) == 0:
                    handed.append("L")
                else:
                    handed.append("R")
                if np.mod(col, 4) < 2:
                    pol_A[p] = 0.0
                else:
                    pol_A[p] = 45.0
                pol_B[p] = 90.0 + pol_A[p]

        pol_A = u.Quantity(pol_A, u.degree)
        pol_B = u.Quantity(pol_B, u.degree)
        layout_A = hex_layout(npix, width, "", "", pol_A, center=hex_cent)
        layout_B = hex_layout(npix, width, "", "", pol_B, center=hex_cent)
    else:
        raise RuntimeError("Unknown wafer packing '{}'".format(wprops["packing"]))

    # Now we go through each pixel and create the orthogonal detectors for
    # each band.
    dets = OrderedDict()

    # chan_per_AMC = cardprops["nchannel"] // cardprops["nAMC"]
    chan_per_AMC = 910
    chan_per_mux = 64
    chan_per_bias = cardprops["nchannel"] // cardprops["nbias"]
    readout_freq_range = np.linspace(4.0, 6.0, chan_per_AMC)
    readout_freq = np.append(readout_freq_range, readout_freq_range)

    doff = 0
    p = 0
    idoff = int(wafer_slot[-2:]) * 10000
    for px in range(npix):
        if px in kill:
            continue
        if npix < 100:
            pxstr = f"{px:02d}"
        else:
            pxstr = f"{px:03d}"
        pstr = f"{p:03d}"
        for b in bands:
            for pl, layout, in zip(
                ["A", "B"],
                [layout_A, layout_B],
            ):
                dprops = OrderedDict()
                dprops["wafer_slot"] = wafer_slot
                dprops["ID"] = idoff + doff
                dprops["pixel"] = pxstr
                dprops["band"] = b
                dprops["fwhm"] = fwhm[b]
                dprops["pol"] = pl

                # Translate the wafer-centric detector quaternion into
                # x/y [mm] position in the UFM basis use by DetMap
                xi, eta, _ = quat_to_xieta(layout[pxstr]["quat"])
                x, y = np.degrees([eta, xi]) / platescale
                # Rotate x and y by 120 degrees
                theta = np.radians(120)
                x, y = np.cos(theta) * x - np.sin(theta) * y, np.sin(theta) * x + np.cos(theta) * y
                dprops["wafer_x_mm"] = x
                dprops["wafer_y_mm"] = y

                # Polarization angle in wafer basis.  This is the gamma angle
                # returned by the layout functions above, less the rotation
                # of the wafer.
                wafer_gamma = layout[pxstr]["gamma"]
                dprops["pol_ang_wafer"] = np.degrees(wafer_gamma - wafer_rot_rad)

                # Physical polarization orientation on wafer is always expressed
                # as 0 or 45.
                dprops["pol_orientation_wafer"] = dprops["pol_ang_wafer"] % 90

                # Polarization angle in focalplane basis.  This is, by
                # definition, equal to the gamma angle computed by the layout
                # functions.  However, we must also account for the extra rotation
                # of the center offset passed to this function.
                if center is not None:
                    dprops["quat"] = qa.mult(center, layout[pxstr]["quat"]).flatten()
                    _, _, temp_gamma = quat_to_xieta(dprops["quat"])
                    dprops["gamma"] = np.degrees(temp_gamma)
                else:
                    dprops["quat"] = layout[pxstr]["quat"].flatten()
                    dprops["gamma"] = np.degrees(wafer_gamma)
                dprops["pol_ang"] = dprops["gamma"]

                if handed is not None:
                    dprops["handed"] = handed[px]
                # Made-up assignment to readout channels
                dprops["card_slot"] = card_slot
                dprops["channel"] = doff
                dprops["AMC"] = doff // chan_per_AMC
                dprops["bias"] = doff // chan_per_bias
                dprops["readout_freq_GHz"] = readout_freq[doff]
                dprops["bondpad"] = doff - (doff // chan_per_mux) * chan_per_mux
                dprops["mux_position"] = doff // chan_per_mux

                dprops["detector_name"] = ""
                dname = "{}_p{}_{}_{}".format(wafer_slot, pxstr, b, pl)
                dets[dname] = dprops
                doff += 1
        p += 1

    return dets


def load_wafer_data(filename, array_name, sub_array_name=None):
    """Load a Det_Info file (like those made by make-det-info-wafer) and return
    a specific array.

    Allows you to replace an array name with a different one (since we
    currently do not have a complete set of array hardware mappings) and
    uses the replacement name pattern to replace bandpass information. If using
    the sub_array_name it assumes array_name is an MF array

    filename(str): path to det_info hdf5 file
    array_name(str): array name to load from the file
    sub_array_name(str): work-around to change the array name while we
        only have some arrays
    """

    rs = read_dataset(filename, array_name)
    rs.keys = loader._filter_items("dets:", rs.keys)

    bp = rs.keys.index("wafer.bandpass")

    if sub_array_name is not None:
        switches = {
            "L" : {"f090" : "f030", "f150" : "f040"},
            "U" : {"f090" : "f230", "f150" : "f290"},
        }
        for i in range(len(rs)):
            row = rs.rows[i]
            rs.rows[i] = tuple([
                r.replace(array_name, sub_array_name) for r in row[:2]
            ]) + row[2:]

            if rs.rows[i][bp] == "NC" or "M" in sub_array_name:
                continue
            if "L" in sub_array_name:
                rs.rows[i] = tuple([
                    rs.rows[i][0].replace(
                        rs.rows[i][bp], switches["L"][rs.rows[i][bp]]
                    ),
                    rs.rows[i][1]
                ]) + rs.rows[i][2:bp] \
                + tuple([switches["L"][rs.rows[i][bp]]]) + rs.rows[i][bp+1:]
            elif "U" in sub_array_name:
                rs.rows[i] = tuple([
                    rs.rows[i][0].replace(
                        rs.rows[i][bp], switches["U"][rs.rows[i][bp]]
                    ),
                    rs.rows[i][1]
                ]) + rs.rows[i][2:bp] + \
                tuple([switches["U"][rs.rows[i][bp]]]) + rs.rows[i][bp + 1:]

    wafer = loader.convert_det_info(
        rs,
        rs["det_id"]

    ).wafer
    return wafer


def load_wafer_detectors(
    hw,
    wafer_slot,
    platescale,
    fwhm,
    det_info_file,
    array_name,
    sub_array_name=None,
    band=None,
    center=None,
    no_darks=False
):

    """Load detector properties for a wafer. Meant to act similarly to
    sim_wafer_detectors

    Given a Hardware configuration, load all detector properties for
    the specified wafer and optionally only the specified band.

    Args:
        hw (Hardware): The hardware properties.
        wafer_slot (str): The wafer slot name.
        platescale (float): The plate scale in degrees / mm.
        fwhm (dict): Dictionary of nominal FWHM values in arcminutes for
            each band.
        det_info_file (str): filename of detector info file
        array_name (str): array to load from file
        sub_array_name (str, optional): array to change to, defaults to
            hw.data["wafer_slots"][slot]["wafer_name"] if not set
        band (str, optional): Optionally only use this band.
        center (array, optional): The quaternion offset of the center.
        no_darks (bool, optional): if true, will not return dark detectors

    Returns:
        (OrderedDict): The properties of all selected detectors.

    """

    # The properties of this wafer
    wprops = hw.data["wafer_slots"][wafer_slot]
    tele = wprops["type"].split("_")[0]

    # The readout card and its properties
    card_slot = wprops["card_slot"]
    cardprops = hw.data["card_slots"][card_slot]

    # The bands
    bands = wprops["bands"]
    if band is not None:
        if band in bands:
            bands = [band]
        else:
            msg = f"band '{band}' not valid for wafer '{wafer_slot}'"
            raise RuntimeError(msg)

    if sub_array_name is None:
        sub_array_name = wprops["wafer_name"]
    wafer = load_wafer_data(det_info_file, array_name, sub_array_name)

    dets = OrderedDict()
    # FIXME: double check this
    # FIXME: this seems REALLY fragile.  Can we do something better?
    poloff = np.degrees(np.nanmin(wafer.angle)) - 22.5

    for i, detname in enumerate(wafer.dets.vals):
        if wafer.dets.vals[i] == 'NO_MATCH':
            continue

        dprops = OrderedDict()
        dprops["wafer_slot"] = wafer_slot
        dprops["ID"] = toast.utils.name_UID(detname)
        dprops["pixel"] = detname.split("_")[-1][:-1]
        if wafer.bandpass[i] != "NC":
            dprops["band"] = f"{tele[:3]}_{wafer.bandpass[i]}"
        else:
            dprops["band"] = "NC"

        if dprops["band"] not in bands and dprops["band"] != "NC":
            continue
        if no_darks and dprops["band"] == "NC":
            continue

        dprops["fwhm"] = fwhm[dprops["band"]] if dprops["band"] in fwhm else np.nan
        dprops["pol"] = wafer.pol[i]

        # This should be the same as the detector gamma angle relative to focalplane
        # coordinates.
        dprops["gamma"] = np.degrees(wafer.angle[i])
        dprops["pol_ang"] = dprops["gamma"]

        # Polarization angle in the wafer coordinate system
        dprops["pol_ang_wafer"] = np.round(np.degrees(wafer.angle[i]) - poloff, 2)

        # This angle seems meaningless for wafers assembled out of rhomboids
        dprops["pol_orientation_wafer"] = dprops["pol_ang_wafer"] % 90

        ## channels aren't assigned until Tunes are made, so just ints
        ## with tunes this will be 512*smurf_band + smurf_channel. not available
        ## with just hardware mapping files.
        dprops["channel"] = i
        ## card slot will be the stream_id name for the wafer slot
        dprops["card_slot"] = f"stream_id_{wafer_slot}"

        ## readout related info
        dprops["bias"] = wafer.bias_line[i]
        dprops["AMC"] = {"N" : 0, "S" : 1}[wafer.coax[i]]
        dprops["readout_freq_GHz"] = wafer.design_freq_mhz[i] / 1000
        dprops["bondpad"] = wafer.bond_pad[i]
        dprops["mux_position"] = wafer.mux_position[i]

        quat = xieta_to_quat(
            np.radians(wafer.det_x[i] * platescale),
            np.radians(wafer.det_y[i] * platescale),
            wafer.angle[i],
        )
        if center is not None:
            dprops["quat"] = qa.mult(center, quat).flatten()
            # Update gamma angle to include the center rotation
            _, _, temp_gamma = quat_to_xieta(dprops["quat"])
            dprops["gamma"] = np.degrees(temp_gamma)
            dprops["pol_ang"] = dprops["gamma"]
        else:
            dprops["quat"] = quat.flatten()
        dprops["detector_name"] = f"{wafer_slot}{detname}"
        dets[dprops["detector_name"]] = dprops

    return dets


def sim_telescope_detectors(hw, tele, tube_slots=None, det_info=None, no_darks=False):
    """Update hardware model with simulated or loaded detector positions.

    Given a Hardware model, generate all detector properties for the specified
    telescope and optionally a subset of optics tube slots (for the LAT).  The
    detector dictionary of the hardware model is updated in place.

    This uses helper functions for focalplane layout from the upstream toast
    package

    Args:
        hw (Hardware): The hardware object to update.
        tele (str): The telescope name.
        tube_slots (list, optional): The optional list of tube slots to include.
        det_info (tuple, optional): Detector info database used to load real
            array hardware.
    Returns:
        None

    """

    zaxis = np.array([0, 0, 1], dtype=np.float64)
    thirty = np.pi / 6.0
    sixty = np.pi / 3.0

    # det_info parameters
    det_info_file = None
    det_info_version = None
    if det_info is not None:
        if len(det_info) < 2:
            raise RuntimeError("Expected at least 2 elements in det_info tuple")
        det_info_file = det_info[0]
        det_info_version = det_info[1]

    # Disable this code path for now, until systematic study or roundtrip
    # unit test is implemented.
    if det_info_file is not None:
        raise NotImplementedError("det_info loading disabled pending unit test")

    # The properties of this telescope
    teleprops = hw.data["telescopes"][tele]
    platescale = teleprops["platescale"]
    fwhm = teleprops["fwhm"]

    # The tubes
    alltubes = teleprops["tube_slots"]
    ntube = len(alltubes)
    if tube_slots is None:
        tube_slots = alltubes
    else:
        for t in tube_slots:
            if t not in alltubes:
                msg = f"Invalid tube_slot '{t}' for telescope '{tele}'"
                raise RuntimeError(msg)

    # All tubes on both SAT and LAT are flipped along both coordinate axes
    # (i.e. rotated 180 degrees) by the reimaging optics.
    reimaging_optics_rot = qa.from_axisangle(zaxis, np.pi)

    alldets = OrderedDict()
    if ntube == 1:
        # This is a SAT.  We have one tube at the center.
        tubeprops = hw.data["tube_slots"][tube_slots[0]]
        waferspace = tubeprops["waferspace"] * platescale

        # This is the per-wafer rotation at it tube location.
        wafer_slot_ang_deg = tubeprops["wafer_slot_angle"]
        wafer_slot_ang_rad = np.radians(wafer_slot_ang_deg)

        # Wafers are arranged in a rotated hexagon shape, and each wafer
        # is individually rotated as well.
        if det_info_file is None:
            tube_center = qa.mult(
                reimaging_optics_rot,
                qa.from_axisangle(zaxis, -thirty),
            )
            extra_wafer_rots = None
        else:
            # FIXME: verify this.
            tube_center = qa.mult(
                reimaging_optics_rot,
                qa.from_axisangle(zaxis, thirty),
            )
            # Additional rotation for each wafer
            # FIXME:  do we understand where this comes from?
            extra_wafer_rots = [
                qa.from_axisangle(zaxis, ang) for ang in [
                    thirty,
                    thirty,
                    -thirty,
                    -3*thirty,
                    5*thirty,
                    3*thirty,
                    thirty,
                ]
            ]

        # Wafer layout, including the overall rotation of the
        # tube center.
        wcenters = hex_layout(
            7,
            2 * waferspace * u.degree,
            "",
            "",
            np.zeros(7, dtype=np.float64) * u.degree,
            center=tube_center,
        )

        # Now rotate each wafer
        centers = np.zeros((7, 4), dtype=np.float64)
        for p, q in wcenters.items():
            windx = int(p)
            wquat = q["quat"]
            wrot = qa.from_axisangle(zaxis, wafer_slot_ang_rad[windx])
            if det_info_file is not None:
                # Add an additional rotation of the wafer
                quat2 = qa.from_axisangle(zaxis, extra_wafer_rots[windx])
                wquat = qa.mult(wquat, quat2)
            centers[windx, :] = qa.mult(wquat, wrot)

        for windx, wafer_slot in enumerate(tubeprops["wafer_slots"]):
            if det_info_file is not None:
                dets = load_wafer_detectors(
                    hw, 
                    wafer_slot, 
                    platescale, 
                    fwhm,
                    det_info_file, 
                    det_info_version, 
                    center=centers[windx],
                    no_darks=no_darks,
                )
            else:
                dets = sim_wafer_detectors(
                    hw, wafer_slot, platescale, fwhm, center=centers[windx]
                )
            alldets.update(dets)
    else:
        # This is the LAT.  We layout detectors for the case of 30
        # degree elevation and no boresight rotation.  Compute the
        # tube centers.

        # Inter-tube spacing
        tubespace = teleprops["tubespace"]

        # Hexagon layout
        tube_quats = hex_layout(
            19,
            4 * (tubespace * platescale) * u.degree,
            "",
            "",
            np.zeros(19, dtype=np.float64) * u.degree,
            center=None,
        )
        tcenters = np.array(
            [q["quat"] for p, q in tube_quats.items()]
        )

        for tindx, tube_slot in enumerate(tube_slots):
            tubeprops = hw.data["tube_slots"][tube_slot]
            waferspace = tubeprops["waferspace"]
            location = tubeprops["toast_hex_pos"]
            wafer_slot_ang_deg = tubeprops["wafer_slot_angle"]
            wafer_slot_ang_rad = np.radians(wafer_slot_ang_deg)

            wradius = 0.5 * (waferspace * platescale * np.pi / 180.0)

            # The wafers within the tube are laid out with the c1-o6 line
            # along the vertical.  Then it is rotated 90 degrees.

            if det_info_file is None:
                qwcenters = [
                    xieta_to_quat(
                        -wradius, wradius / np.sqrt(3.0), 0.0
                    ),
                    xieta_to_quat(
                        wradius, wradius / np.sqrt(3.0), 0.0
                    ),
                    xieta_to_quat(
                        0.0, -2.0 * wradius / np.sqrt(3.0), 0.0
                    ),
                ]
            else:
                # FIXME: again, det_info case needs re-testing.
                qwcenters = [
                    xieta_to_quat(
                        -wradius / np.sqrt(3.0), wradius, 0.0
                    ),
                    xieta_to_quat(
                        2.0 * wradius / np.sqrt(3.0), 0.0, 0.0
                    ),
                    xieta_to_quat(
                        -wradius / np.sqrt(3.0), -wradius, 0.0
                    ),
                ]

            centers = list()
            for windx, qwc in enumerate(qwcenters):
                centers.append(
                    qa.mult(
                        qa.mult(
                            tcenters[location],
                            qa.mult(reimaging_optics_rot, qwc),
                        ),
                        qa.from_axisangle(zaxis, wafer_slot_ang_rad[windx] - sixty),
                    )
                )

            for windx, wafer_slot in enumerate(tubeprops["wafer_slots"]):
                if det_info_file is not None:
                    dets = load_wafer_detectors(
                        hw,
                        wafer_slot,
                        platescale,
                        fwhm,
                        det_info_file,
                        det_info_version,
                        center=centers[windx],
                        no_darks=no_darks,
                    )
                else:
                    dets = sim_wafer_detectors(
                        hw, wafer_slot, platescale, fwhm, center=centers[windx]
                    )
                # Accumulate to the total
                alldets.update(dets)

        # Rotate the focalplane to the nominal horizontal orientation.
        fp_rot = qa.from_axisangle(zaxis, 0.5 * np.pi)
        for d, props in alldets.items():
            props["quat"] = qa.mult(fp_rot, props["quat"])
            # Update gamma angle to include this rotation
            _, _, temp_gamma = quat_to_xieta(props["quat"])
            props["gamma"] = np.degrees(temp_gamma)
            props["pol_ang"] = props["gamma"]

    if "detectors" in hw.data:
        hw.data["detectors"].update(alldets)
    else:
        hw.data["detectors"] = alldets



import plotly
from plotly.subplots import make_subplots
import matplotlib.colors as mcolors
from plotly.colors import DEFAULT_PLOTLY_COLORS
import plotly.graph_objects as go
import numpy as np
import pandas as pd
from dataclasses import dataclass, field
import plotly.express as px
import datetime as dt
from copy import deepcopy
import ast
from collections import defaultdict

from typing import List, Tuple, TYPE_CHECKING, Dict, Any, Optional

from .report_data import ReportData, Footprint, obs_list_to_arr

band_colors = [
    "#D55E00",
    "#56B4E9",
]

def get_discrete_distinct_colors(n, reverse=False):
    base_colors = ['#4477AA', '#EE6677', '#228833', '#CCBB44', '#66CCEE', '#AA3377', '#BBBBBB']
    n_base = len(base_colors)

    if n <= n_base:
        # Use subset with fixed endpoints
        colors = [base_colors[0]] + base_colors[1:-1][:n - 2] + [base_colors[-1]]
    else:
        # Interpolate between base colors
        base_rgb = [mcolors.to_rgb(c) for c in base_colors]
        interpolated = np.linspace(0, n_base - 1, n)
        colors = []
        for i in interpolated:
            i_low = int(np.floor(i))
            i_high = min(i_low + 1, n_base - 1)
            t = i - i_low
            rgb = (1 - t) * np.array(base_rgb[i_low]) + t * np.array(base_rgb[i_high])
            colors.append(mcolors.to_hex(rgb))

    if reverse:
        colors = list(reversed(colors))

    return colors


@dataclass
class Colors:
    names: List[str]
    reverse: bool = False
    colormap: Dict[str, str] = field(init=False)

    def __post_init__(self):
        n = len(self.names)
        discrete_colors = get_discrete_distinct_colors(n, self.reverse)
        self.colormap = {name: color for name, color in zip(self.names, discrete_colors)}

    def __getitem__(self, name: str) -> str:
        return self.colormap.get(name, "#CCCCCC")

    def __repr__(self):
        return f"Colors({self.colormap})"


@dataclass
class ObsEfficiencyPlots:
    pie: go.Figure
    heatmap: go.Figure


def wafer_obs_efficiency(d: ReportData, nsegs=2000) -> ObsEfficiencyPlots:
    if d.cfg.platform  == "lat":
        tube_slots = ["c1", "i1", "i3", "i4", "i5", "i6"]
        wafer_slots = ["ws0", "ws1", "ws2"]
        wafers = [f"{x}_{y}" for x in tube_slots for y in wafer_slots]
    elif d.cfg.platform in ["satp1", "satp2", "satp3"]:
        wafers = [f"ws{i} " for i in range(7)]
    else:
        raise ValueError(f"Uknown platform {d.cfg.platform}")

    nwafers = len(wafers)

    times = pd.date_range(d.cfg.start_time, d.cfg.stop_time, nsegs).to_pydatetime()
    tstamps= np.array([t.timestamp() for t in times])

    obs_types = ["cmb", "obs", "cal", "oper"] + d.cfg.cal_targets + ["idle"]
    obs_values = {k: i for i, k in enumerate(obs_types)}

    data = np.ones((nwafers, nsegs), dtype=int) * (len(obs_types) - 1)

    for o in d.obs_list:
        m = np.logical_and.reduce([tstamps > o.start_time, tstamps < o.stop_time])
        wafer_slots_list = o.wafer_slots_list.split(",")
        for wafer_slot in wafer_slots_list:
            if d.cfg.platform == "lat":
                idx = np.where(np.array(wafers) == o.obs_tube_slot + "_" + wafer_slot)[0][0]
            else:
                idx = int(wafer_slot.strip()[-1])
            if o.obs_type == "obs":
                if o.obs_subtype == "cmb":
                    data[idx][m] = obs_values[o.obs_subtype]
                elif o.obs_subtype == "cal":
                    matches = [item for item in o.obs_tags.split(',') if item in d.cfg.cal_targets]
                    if matches:
                        data[idx][m] = obs_values[matches[0]]
                    else:
                        data[idx][m] = obs_values[o.obs_subtype]
                else:
                    data[idx][m] = obs_values[o.obs_type]
            else:
                data[idx][m] = obs_values[o.obs_type]

    reverse = False
    if (data == (len(obs_types) - 1)).all():
        reverse = True

    # Compile data for pie chart
    unique_vals, counts = np.unique(data, return_counts=True)
    percentages = counts / counts.sum() * 100
    labels = [obs_types[i] for i in unique_vals]
    COLORS = Colors(names=labels, reverse=reverse)
    colorscale: List[Tuple[float, str]] = []
    ntypes = len(labels)
    for i, t in enumerate(labels):
        colorscale.extend([(i / ntypes, COLORS[t]), ((i + 1) / ntypes, COLORS[t])])
    pie_colors = [COLORS[t] for t in labels]
    colorbar=dict(tickvals=list(range(len(labels))), ticktext=labels,)

    unique_sorted = np.sort(np.unique(data))
    mapping = {v: i for i, v in enumerate(unique_sorted)}
    vectorized_map = np.vectorize(mapping.get)
    data = vectorized_map(data)

    value_to_label = {i: lbl for i, lbl in enumerate(labels)}
    text = np.vectorize(value_to_label.get)(data)

    ys = wafers

    hover_text = []
    for yi, yval in enumerate(ys):
        row = []
        for xi, xval in enumerate(times):
            label = text[yi, xi]
            row.append(f"x: {xval}<br>y: {yval}<br>z: {label}")
        hover_text.append(row)

    heatmap = go.Figure(
        data=go.Heatmap(
            z=data,
            zmin=0,
            zmax=len(labels) - 1,
            x=times,
            y=ys,
            text=hover_text,
            hoverinfo="text",
            colorscale=colorscale,
            colorbar=dict(tickvals=list(range(len(labels))), ticktext=labels,),
            ygap=1,
        ),
    )
    heatmap.update_layout(margin=dict(l=0, r=0, t=0, b=0), height=300)

    pie = go.Figure(
        data=go.Pie(
            labels=labels,
            values=percentages,
            textinfo="label+percent",
        ),
    )
    pie.update_traces(marker=dict(colors=pie_colors))
    pie.update_layout(margin=dict(l=0, r=0, t=0, b=0), height=300)

    return ObsEfficiencyPlots(pie=pie, heatmap=heatmap)


def yield_vs_pwv(d: "ReportData", longterm_data: Optional["ReportData"] = None) -> go.Figure:
    fig = go.Figure()

    if longterm_data is not None:
        band_data = defaultdict(lambda: {"pwv": [], "yield": []})

        for obs in longterm_data.obs_list:
            if not obs.num_valid_dets:
                continue

            band_yields = ast.literal_eval(obs.num_valid_dets)

            for band, val in band_yields.items():
                if np.isfinite(val):
                    band_data[band]["pwv"].append(obs.pwv)
                    band_data[band]["yield"].append(val)

        band_index = 0
        for band, vals in band_data.items():
            fig.add_trace(
                go.Histogram2dContour(
                    x=vals["pwv"],
                    y=vals["yield"],
                    colorscale=[[0, band_colors[band_index]], [1, band_colors[band_index]]],
                    contours_coloring="lines",
                    showscale=False,
                    opacity=0.7,
                    name=f"longterm {band}"
                )
            )
            band_index +=1

    band_index = 0
    band_data = defaultdict(lambda: {'x': [], 'y': [], 'hover': []})

    for obs in d.obs_list:
        ts = dt.datetime.fromtimestamp(obs.start_time, tz=dt.timezone.utc).isoformat()
        if obs.num_valid_dets:
            for band, val in ast.literal_eval(obs.num_valid_dets).items():
                if not np.isfinite(val):
                    continue
                band_data[band]['x'].append(obs.pwv)
                band_data[band]['y'].append(val)
                band_data[band]['hover'].append(obs.obs_id)

    for band, data in band_data.items():
        color = band_colors[band_index % len(band_colors)]
        fig.add_trace(
            go.Scatter(
                x=data['x'],
                y=data['y'],
                mode="markers",
                name=f"{band}",
                marker=dict(color=color),
                hovertext=data['hover'],
            )
        )
        band_index += 1

    fig.update_layout(
        xaxis=dict(title="PWV"),
        yaxis=dict(title="Valid Dets"),
        height=500,
        margin=dict(l=0, r=0, t=0, b=0),
    )
    return fig



def pwv_and_yield_vs_time(d: "ReportData") -> go.Figure:
    fig = make_subplots(specs=[[{"secondary_y": True}]])

    # Collect yields per band
    yields_by_band = defaultdict(list)
    times_by_band = defaultdict(list)
    hover_by_band = defaultdict(list)

    for obs in d.obs_list:
        obs_time = dt.datetime.fromtimestamp(obs.start_time, tz=dt.timezone.utc)
        if obs.num_valid_dets:
            for band, val in ast.literal_eval(obs.num_valid_dets).items():
                times_by_band[band].append(obs_time)
                yields_by_band[band].append(val)
                hover_by_band[band].append(obs.obs_id)

    band_index = 0

    for band in sorted(yields_by_band):
        color = band_colors[band_index % len(band_colors)]
        fig.add_trace(
            go.Scatter(
                x=times_by_band[band],
                y=yields_by_band[band],
                mode='markers',
                name=f"Valid Dets ({band})",
                marker=dict(color=color),
                hovertext=hover_by_band[band],
            ),
            secondary_y=False,
        )
        band_index += 1

    # Add PWV trace
    ds_factor = 10
    pwvs = deepcopy(d.pwv[1][::ds_factor])
    pwvs[(pwvs > 4) | (pwvs < .1)] = np.nan
    ts = [dt.datetime.fromtimestamp(t, tz=dt.timezone.utc) for t in d.pwv[0][::ds_factor]]

    fig.add_trace(
        go.Scatter(
            x=ts,
            y=pwvs,
            mode='lines',
            name="PWV",
            marker=dict(color="#E69F00"),
            opacity=0.5,
        ),
        secondary_y=True
    )

    # Axis labels and layout
    fig.update_yaxes(title_text='Num Valid Dets', secondary_y=False)
    fig.update_yaxes(title_text='PWV', secondary_y=True)
    fig.update_layout(
        margin={k: 0 for k in ['l', 'r', 't', 'b']},
        height=500
    )

    return fig


@dataclass
class SourceFootprintPlots:
    focalplane: go.Figure
    table: go.Figure


def source_footprints(d: "ReportData") -> go.Figure:

    if d.cfg.platform in ["satp1", "satp2", "satp3"]:
        wafer_rad = 0.10488
        wafer_sep = 0.219911
        pie_width = 0.25

        angs = np.arange(-np.pi / 2, 3 * np.pi / 2, np.pi / 3)
        x0s = [0] + [np.cos(a) * wafer_sep for a in angs]
        y0s = [0] + [np.sin(a) * wafer_sep for a in angs]

        wafers = ['ws0', 'ws1', 'ws6', 'ws5', 'ws4', 'ws3', 'ws2']

        wafer_centers = {}
        for i, wafer in enumerate(wafers):
            wafer_centers[wafer] = (x0s[i], y0s[i])

    elif d.cfg.platform == "lat":
        wafer_rad = 0.005236
        pie_width = 0.125
        x0s = [-0.00647517,  0.00316777,  0.00316777, -0.03335673, -0.02370855,
               -0.02371379,  0.02070833,  0.03023957,  0.03025179,  0.0204762 ,
                0.03025005,  0.03023957, -0.00637918,  0.00327947,  0.00325853,
               -0.03330437, -0.02369634, -0.02370855]

        y0s = [ 0.        ,  0.00560425, -0.00560425, -0.01579872, -0.00995536,
               -0.02117608, -0.01556659, -0.0099571 , -0.02117957,  0.01579872,
                0.02117957,  0.0099571 ,  0.03112446,  0.03673045,  0.02551671,
                0.01556834,  0.02117608,  0.01021716]

        wafers = ['c1_ws0', 'c1_ws1', 'c1_ws2',
                  'i1_ws0', 'i1_ws1', 'i1_ws2',
                  'i3_ws0', 'i3_ws1', 'i3_ws2',
                  'i4_ws0', 'i4_ws1', 'i4_ws2',
                  'i5_ws0', 'i5_ws1', 'i5_ws2',
                  'i6_ws0', 'i6_ws1', 'i6_ws2']

        wafer_centers = {}
        for i, wafer in enumerate(wafers):
            wafer_centers[wafer] = (x0s[i], y0s[i])

    def hex(x0, y0):
        angs = np.linspace(0, 2 * np.pi, 7)
        pts = np.array(
            [
                x0 + wafer_rad * np.cos(angs),
                y0 + wafer_rad * np.sin(angs),
            ]
        )
        return pts

    xs = [x for x, y in wafer_centers.values()]
    ys = [y for x, y in wafer_centers.values()]

    x_min, x_max = min(xs), max(xs)
    y_min, y_max = min(ys), max(ys)
    x_span = x_max - x_min
    y_span = y_max - y_min
    if x_span > y_span:
        y_mid = (y_max + y_min) / 2
        y_min = y_mid - x_span / 2
        y_max = y_mid + x_span / 2
    else:
        x_mid = (x_max + x_min) / 2
        x_min = x_mid - y_span / 2
        x_max = x_mid + y_span / 2
    x_margin = x_span * 0.05
    y_margin = x_span * 0.05

    x_range = [x_min - 1.1*wafer_rad, x_max + 1.1*wafer_rad]
    y_range = [y_min - 1.1*wafer_rad, y_max + 1.1*wafer_rad]

    def get_normalized_positions(positions, x_range, y_range, fig_width, fig_height):
        x0, x1 = x_range
        y0, y1 = y_range
        x_span = x1 - x0
        y_span = y1 - y0

        data_aspect = x_span / y_span
        fig_aspect = fig_width / fig_height

        norm_positions = {}

        if data_aspect > fig_aspect:
            # Y is scaled (shorter)
            scale = fig_aspect / data_aspect
            y_offset = (1 - scale) / 2
            for key, (x, y) in positions.items():
                x_paper = (x - x0) / x_span
                y_paper = y_offset + scale * (y - y0) / y_span
                norm_positions[key] = (x_paper, y_paper)
        else:
            # X is scaled (shorter)
            scale = data_aspect / fig_aspect
            x_offset = (1 - scale) / 2
            for key, (x, y) in positions.items():
                x_paper = x_offset + scale * (x - x0) / x_span
                y_paper = (y - y0) / y_span
                norm_positions[key] = (x_paper, y_paper)

        return norm_positions

    norm_positions = get_normalized_positions(wafer_centers, x_range, y_range, 700, 700)

    def normalize_values(values, target_total=100):
        total = sum(values)
        if total == 0:
            return values  # leave empty pie alone
        return [v * target_total / total for v in values]

    fig = go.Figure()

    for i, (x0, y0) in enumerate(zip(x0s, y0s)):
        pts = hex(x0, y0)
        fig.add_trace(
            go.Scatter(
                x=pts[0],
                y=pts[1],
                fill="toself",
                fillcolor="grey",
                opacity=0.3,
                mode="lines",
                marker=dict(color="grey"),
                showlegend=False,
                hoverinfo="skip",
            )
        )

        fig.add_annotation(
            x=x0,
            y=y0 + 0.9*wafer_rad,
            text=wafers[i],
            showarrow=False,
            font=dict(size=12),
            xanchor='center',
            yanchor='bottom',
            xref='x',
            yref='y'
        )

    ntargets = 0
    target_colors = {}
    if d.source_footprints is not None:
        wafer_map = defaultdict(list)
        for fp in d.source_footprints:
            wafer_map[fp.wafer].append(fp)

        if wafer_map:
            for wafer, group in wafer_map.items():
                x, y = norm_positions[wafer]
                labels = [fp.source for fp in group]
                values = [fp.count for fp in group]
                hovertext = [
                    f"Wafer: {fp.wafer}<br>Source: {fp.source}<br>Count: {fp.count}<br>ObsIDs:<br>" + "<br>".join(fp.obsids)
                    for fp in group
                ]

                x0 = x - pie_width / 2
                x1 = x + pie_width / 2
                y0 = y - pie_width / 2
                y1 = y + pie_width / 2

                fig.add_trace(
                    go.Pie(
                        labels=labels,
                        values=normalize_values(values),
                        name=wafer,
                        domain=dict(x=[x0, x1], y=[y0, y1]),
                        scalegroup='wafer_pies',
                        hoverinfo='text',
                        hovertext=hovertext
                    )
                )

    fig.add_trace(go.Scatter(
        x=[x_min, x_max],
        y=[y_min, y_max],
        mode='markers',
        marker=dict(opacity=0),
        showlegend=False,
        hoverinfo='skip'
    ))

    fig.update_layout(
        xaxis=dict(
            range=x_range,
            title='Xi (rad)',
            showgrid=True,
            zeroline=True,
            fixedrange=True
        ),
        yaxis=dict(
            range=y_range,
            title='Eta (rad)',
            showgrid=True,
            zeroline=True,
            fixedrange=True
        ),
        width=800,
        height=800,
        margin=dict(t=50, b=50, l=50, r=50)
    )

    def make_obsid_links(obsids):
        if not obsids:
            return ""
        links = [
            f'<a href="{d.cfg.site_url}/site-pipeline/{d.cfg.platform}/preprocess/{obsid.split("_")[1][:5]}/{obsid}/" target="_blank">{obsid}</a>'
            for obsid in obsids
        ]
        return "<br>".join(links) + ("<br>" if len(links) == 1 else "")

    if d.source_footprints is not None:
        wafers = sorted(set(fp.wafer for fp in d.source_footprints))
        sources = sorted(set(fp.source for fp in d.source_footprints))

        lookup = {
            (fp.wafer, fp.source): fp.obsids
            for fp in d.source_footprints
        }

        wafer_col = wafers
        source_cols = []
        for source in sources:
            col = []
            for wafer in wafers:
                obsids = lookup.get((wafer, source), [])
                col.append(make_obsid_links(obsids))
            source_cols.append(col)
    else:
        wafer_col = []
        source_cols = []
        sources = []

    table_columns = [wafer_col] + source_cols
    header_labels = ["Wafers"] + sources

    table = go.Figure(data=[go.Table(
        header=dict(
            values=header_labels,
            fill_color='lightgrey',
            align='left',
            font=dict(size=12),
            line_color='black'
        ),
        cells=dict(
            values=table_columns,
            fill_color='white',
            align='left',
            line_color='black',
            height=30
        )
    )])

    table.update_layout(
        width=500,
        height=400 + 30 * len(wafers),
        margin=dict(l=20, r=20, t=20, b=20),
        dragmode=False
    )

    return SourceFootprintPlots(focalplane=fig, table=table)



# Copyright (c) 2023-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Timestream processing filters.
"""

import numpy as np
from astropy import units as u
import toast
import toast.ops

from .. import ops as so_ops
from .job import workflow_timer


def setup_readout_filter(operators):
    """Add commandline args and operators for readout filter.
    Args:
        operators (list):  The list of operators to extend.
    Returns:
        None
    """
    operators.append(
        so_ops.ReadoutFilter(name="readout_filter", enabled=False)
    )


@workflow_timer
def apply_readout_filter(job, otherargs, runargs, data):
    """Apply readout filter.
    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.
    Returns:
        None
    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.readout_filter.enabled:
        job_ops.readout_filter.apply(data)


def setup_deconvolve_detector_timeconstant(operators):
    """Add commandline args and operators for timeconstant convolution.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(
        toast.ops.TimeConstant(
            name="deconvolve_time_constant",
            deconvolve=True,
            enabled=False,
        )
    )


@workflow_timer
def deconvolve_detector_timeconstant(job, otherargs, runargs, data):
    """Deconvolve the detector timeconstants.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.deconvolve_time_constant.enabled:
        job_ops.deconvolve_time_constant.apply(data)


def setup_filter_hwpss(operators):
    """Add commandline args and operators for HWPSS filters.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(toast.ops.HWPFilter(name="hwpfilter", enabled=False))


@workflow_timer
def filter_hwpss(job, otherargs, runargs, data):
    """Filter HWP synchronous signal.

    This filters the HWPSS, as opposed to solving for it in the mapmaking
    with a template.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.hwpfilter.enabled:
        job_ops.hwpfilter.apply(data)


def setup_filter_hwpss_model(operators):
    """Add commandline args and operators for HWPSS modeling.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(
        toast.ops.HWPSynchronousModel(name="hwpss_model", enabled=False)
    )


@workflow_timer
def filter_hwpss_model(job, otherargs, runargs, data):
    """Model HWPSS and subtract.

    This builds a model for the HWPSS, optionally subtracts it, and
    optionally computes relative calibration factors from the 2f signal.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.hwpss_model.enabled:
        job_ops.hwpss_model.apply(data)


def setup_filter_ground(operators):
    """Add commandline args and operators for ground filters.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(toast.ops.GroundFilter(name="groundfilter", enabled=False))


@workflow_timer
def filter_ground(job, otherargs, runargs, data):
    """Filter Azimuth synchronous signal.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.groundfilter.enabled:
        job_ops.groundfilter.apply(data)


def setup_filter_poly1d(operators):
    """Add commandline args and operators for 1D polynomial filters.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(toast.ops.PolyFilter(name="polyfilter1D", enabled=False))


@workflow_timer
def filter_poly1d(job, otherargs, runargs, data):
    """Filter scans with 1D polynomial.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.polyfilter1D.enabled:
        job_ops.polyfilter1D.apply(data)


def setup_filter_poly2d(operators):
    """Add commandline args and operators for 2D polynomial filters.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(toast.ops.PolyFilter2D(name="polyfilter2D", enabled=False))


@workflow_timer
def filter_poly2d(job, otherargs, runargs, data):
    """Filter data with a 2D polynomial.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.polyfilter2D.enabled:
        job_ops.polyfilter2D.apply(data)


def setup_filter_common_mode(operators):
    """Add commandline args and operators for common mode filters.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(
        toast.ops.CommonModeFilter(
            name="common_mode_filter",
            regress=True,
            enabled=False,
        )
    )


@workflow_timer
def filter_common_mode(job, otherargs, runargs, data):
    """Filter data to remove common modes across the detectors.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.common_mode_filter.enabled:
        job_ops.common_mode_filter.apply(data)
